from skidy.parser import robot_from_yaml
import time
import os
import matplotlib.pyplot as plt
import numpy as np

def benchmark(parallel: bool=True,simplify: bool=True, cse: bool=False, all: bool=False, save: bool=True) -> None:
    # run through all options
    if all:
        for parallel in [True, False]:
            for simplify in [True, False]:
                for cse in [True, False]:
                    benchmark(parallel,simplify,cse)
        return
    
    os.chdir(os.path.dirname(__file__))

    r1 = robot_from_yaml("./r.yaml")
    r2 = robot_from_yaml("./rr.yaml")
    r3 = robot_from_yaml("./rrr.yaml")
    r4 = robot_from_yaml("./rrrr.yaml")


    startr1 = time.time()
    r1.closed_form_kinematics_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    kinr1 = time.time()
    r1.closed_form_inv_dyn_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    dynr1 =  time.time()
    r1.generate_code(python=True,name="r1")
    genr1 = time.time()

    startr2 = time.time()
    r2.closed_form_kinematics_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    kinr2 = time.time()
    r2.closed_form_inv_dyn_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    dynr2 =  time.time()
    r2.generate_code(python=True,name="r2")
    genr2 = time.time()

    startr3 = time.time()
    r3.closed_form_kinematics_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    kinr3 = time.time()
    r3.closed_form_inv_dyn_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    dynr3 =  time.time()
    r3.generate_code(python=True,name="r3")
    genr3 = time.time()

    startr4 = time.time()
    r4.closed_form_kinematics_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    kinr4 = time.time()
    r4.closed_form_inv_dyn_body_fixed(simplify=simplify, cse=cse, parallel=parallel)
    dynr4 =  time.time()
    r4.generate_code(python=True,name="r4")
    genr4 = time.time()

    barkin = [kinr1-startr1, kinr2-startr2, kinr3-startr3,kinr4-startr4]
    bardyn = [dynr1-kinr1, dynr2-kinr2, dynr3-kinr3, dynr4-kinr4]
    bargen = [genr1-dynr1, genr2-dynr2, genr3-dynr3, genr4-dynr4]
    cumtime = [genr1-startr1,genr2-startr2,genr3-startr3,genr4-startr4]
    
    r = [1,2,3,4]
    
    plt.figure()
    
    plt.bar(r, barkin, color='#7f6d5f', edgecolor='white', width=.8, label="forward kinematics")
    plt.bar(r, bardyn, bottom=barkin, color='#557f2d', edgecolor='white', width=.8, label="inverse dynamics")
    plt.bar(r, bargen, bottom=np.add(barkin,bardyn).tolist(), color='#2d7f5e', edgecolor='white', width=.8, label="code generation")

    for i in range(len(r)):
        plt.text(r[i]-0.2, cumtime[i]+0.2, "%.02f s"%cumtime[i])
    
    plt.xticks(r, r)
    plt.xlabel("degrees of freedom")
    plt.ylabel("time in s")

    plt.legend()

    title = ""
    if parallel:
        title += "parallel; "
    else:
        title += "serial; "
    if simplify:
        title += "with simplification; "
    else:
        title += "without simplification; "
    if cse:    
        title += "with cse"
    else:
        title += "without cse"

    plt.title(title)
    name = title.replace(";","").replace(" ","_") + ".png"
    if save:
        plt.savefig(name)
    else:
        plt.show()


if __name__ == "__main__":
    benchmark(all=True)    


from __future__ import annotations

from sympy import Matrix, Identity, symbols, sin, cos, zeros, MutableDenseMatrix, Expr
import numpy as np
from numpy.typing import ArrayLike

def generalized_vectors(
    DOF: int, startindex: int=0, derivatives: bool=False
    ) -> tuple[MutableDenseMatrix, MutableDenseMatrix, MutableDenseMatrix]:
    """Generate symbolic generalized vectors q, qd and q2d and optional
    q3d and q4d.
    
    The symbols are named as follows:
        
        q0, q1, ....., qi for joint positions.
        dq0, dq1, ....., dqi for joint velocities.
        ddq0, ddq1, ....., ddqi for joint accelerations.
        dddq0, dddq1, ....., dddqi for joint jerks.
        ddddq0, ddddq1, ....., ddddqi for joint jounces.
        

    Args:
        DOF (int): Degrees of freedom.
        startindex (int, optional): Index of first joint. Defaults to 0.
        derivatives (bool, optional): return jerks and jounces too. 
            Defaults to False. 

    Returns:
        tuple(sympy.Matrix): 
            Generalized vectors: (q,dq,d2q) if derivative == False
            else: (q,dq,d2q,q3d,q4d) 
    """
    if DOF > 1:
        q = Matrix(symbols(" ".join(f"q{i}" for i in range(startindex,startindex+DOF))))
        qd = Matrix(symbols(" ".join(f"dq{i}" for i in range(startindex,startindex+DOF))))
        q2d = Matrix(symbols(" ".join(f"ddq{i}" for i in range(startindex,startindex+DOF))))
        q3d = Matrix(symbols(" ".join(f"dddq{i}" for i in range(startindex,startindex+DOF))))
        q4d = Matrix(symbols(" ".join(f"ddddq{i}" for i in range(startindex,startindex+DOF))))
    else:
        q = Matrix([symbols(" ".join(f"q{i}" for i in range(startindex,startindex+DOF)))])
        qd = Matrix([symbols(" ".join(f"dq{i}" for i in range(startindex,startindex+DOF)))])
        q2d = Matrix([symbols(" ".join(f"ddq{i}" for i in range(startindex,startindex+DOF)))])
        q3d = Matrix([symbols(" ".join(f"dddq{i}" for i in range(startindex,startindex+DOF)))])
        q4d = Matrix([symbols(" ".join(f"ddddq{i}" for i in range(startindex,startindex+DOF)))])
    if derivatives:
        return q, qd, q2d, q3d, q4d
    else: 
        return q, qd, q2d

def joint_screw(axis: list, vec: list=[0,0,0], revolute: bool=True) -> MutableDenseMatrix:
    """Get joint screw coordinates from joint axis and vector to joint.

    Args:
        axis (list): 
            Joint axis w.r.t. inertial frame.
        vec (list, optional): 
            Vector to joint axis from inertial frame for revolute joint. 
            Defaults to [0,0,0].
        revolute (bool, optional): 
            Revolute (True) or prismatic (False) joint. 
            Defaults to True.

    Returns:
        sympy.Matrix: joint screw coordinates.
    """
    e = Matrix(axis)
        
    if revolute:
        y = Matrix(vec)
        return Matrix([e, y.cross(e)])
    else:
        return Matrix([0,0,0,*e])
        

def symbolic_inertia_matrix(
    index: int | str="", pointmass: bool=False) -> MutableDenseMatrix:
    """Create 3 x 3 symbolic inertia matrix with auto generated variable names.

    Args:
        index (int or str): 
            postfix for variable name. Defaults to "".
        pointmass (bool): 
            Inertial matrix = I * Identity. Default to False.
    
    Returns:
        sympy.Matrix: Inertia matrix (3,3)
    """
    if pointmass:
        I = symbols(f"I{index}", real=1, constant=1)
        return I*Identity(3)
    Ixx, Iyy, Izz, Ixy, Ixz, Iyz = symbols(
        f"Ixx{index} Iyy{index} Izz{index} Ixy{index} Ixz{index} Iyz{index}")
    
    I = Matrix([[Ixx, Ixy, Ixz],
                [Ixy, Iyy, Iyz],
                [Ixz, Iyz, Izz]])
    return I

def SE3AdjInvMatrix(C: MutableDenseMatrix) -> MutableDenseMatrix:
    """Compute Inverse of (6x6) Adjoint matrix for SE(3)

    Args:
        C (sympy.Matrix): SE(3) Pose.

    Returns:
        sympy.Matrix: Inverse of (6x6) Adjoint matrix
    """
    AdInv = Matrix([[C[0, 0], C[1, 0], C[2, 0], 0, 0, 0],
                    [C[0, 1], C[1, 1], C[2, 1], 0, 0, 0],
                    [C[0, 2], C[1, 2], C[2, 2], 0, 0, 0],
                    [-C[2, 3]*C[1, 0]+C[1, 3]*C[2, 0], 
                        C[2, 3]*C[0, 0]-C[0, 3]*C[2, 0],
                        (-C[1, 3])*C[0, 0]+C[0, 3]*C[1, 0], 
                        C[0, 0], C[1, 0], C[2, 0]],
                    [-C[2, 3]*C[1, 1]+C[1, 3]*C[2, 1], 
                        C[2, 3]*C[0, 1]-C[0, 3]*C[2, 1],
                        (-C[1, 3])*C[0, 1]+C[0, 3]*C[1, 1], 
                        C[0, 1], C[1, 1], C[2, 1]],
                    [-C[2, 3]*C[1, 2]+C[1, 3]*C[2, 2], 
                        C[2, 3]*C[0, 2]-C[0, 3]*C[2, 2],
                        (-C[1, 3])*C[0, 2]+C[0, 3]*C[1, 2], 
                        C[0, 2], C[1, 2], C[2, 2]]])
    return AdInv

def SE3AdjMatrix(C: MutableDenseMatrix) -> MutableDenseMatrix:
    """Compute (6x6) Adjoint matrix for SE(3)

    Args:
        C (sympy.Matrix): SE(3) Pose.

    Returns:
    sympy.Matrix: (6x6) Adjoint matrix
    """
    Ad = Matrix([[C[0, 0], C[0, 1], C[0, 2], 0, 0, 0],
                    [C[1, 0], C[1, 1], C[1, 2], 0, 0, 0],
                    [C[2, 0], C[2, 1], C[2, 2], 0, 0, 0],
                    [-C[2, 3]*C[1, 0]+C[1, 3]*C[2, 0], 
                    -C[2, 3]*C[1, 1]+C[1, 3]*C[2, 1],
                    -C[2, 3]*C[1, 2]+C[1, 3]*C[2, 2], 
                    C[0, 0], C[0, 1], C[0, 2]],
                    [C[2, 3]*C[0, 0]-C[0, 3]*C[2, 0],  
                    C[2, 3]*C[0, 1]-C[0, 3]*C[2, 1],
                    C[2, 3]*C[0, 2]-C[0, 3]*C[2, 2], 
                    C[1, 0], C[1, 1], C[1, 2]],
                    [-C[1, 3]*C[0, 0]+C[0, 3]*C[1, 0], 
                    -C[1, 3]*C[0, 1]+C[0, 3]*C[1, 1],
                    -C[1, 3]*C[0, 2]+C[0, 3]*C[1, 2], 
                    C[2, 0], C[2, 1], C[2, 2]]])
    return Ad

def SE3adMatrix(X: MutableDenseMatrix) -> MutableDenseMatrix:
    """Compute (6x6) adjoint matrix for SE(3) 
        - also known as spatial cross product in the literature.

    Args:
        X (sympy.Matrix): (6x1) spatial vector.

    Returns:
        sympy.Matrix: (6x6) adjoint matrix
    """
    ad = Matrix([[0, -X[2, 0], X[1, 0], 0, 0, 0],
                 [X[2, 0], 0, -X[0, 0], 0, 0, 0],
                 [-X[1, 0], X[0, 0], 0, 0, 0, 0],
                 [0, -X[5, 0], X[4, 0], 0, -X[2, 0], X[1, 0]],
                 [X[5, 0], 0, -X[3, 0], X[2, 0], 0, -X[0, 0]],
                 [-X[4, 0], X[3, 0], 0, -X[1, 0], X[0, 0], 0]])
    return ad

def SE3Exp(XX: MutableDenseMatrix, t: float | Expr) -> MutableDenseMatrix:
    """compute exponential mapping for SE(3).

    Args:
        XX (sympy.Matrix): (6,1) spatial vector.
        t (sympy.Expr): rotation angle.

    Returns:
        sympy.Matrix: (4,4) SE(3) Pose.
    """
    X = XX.T
    xi = Matrix(X[0:3])
    eta = Matrix(X[3:6])
    xihat = Matrix([[0, -X[2], X[1]],
                    [X[2], 0, -X[0]],
                    [-X[1], X[0], 0]])
    R = Matrix(Identity(3)) + sin(t)*xihat + (1-cos(t))*(xihat*xihat)
    if xi == zeros(3, 1):
        p = eta * t
    else:
        p = (Matrix(Identity(3))-R)*(xihat*eta) + xi*(xi.T*eta)*t
    C = R.row_join(p).col_join(Matrix([0, 0, 0, 1]).T)
    return C

def SE3Inv(C: MutableDenseMatrix) -> MutableDenseMatrix:
    """Compute analytical inverse of exponential mapping for SE(3).

    Args:
        C (sympy.Matrix): (4,4) SE(3) Pose.

    Returns:
        sympy.Matrix: (4,4) Inverse of SE(3) Pose.
    """
    CInv = Matrix([[C[0, 0], C[1, 0], C[2, 0], 
                    -C[0, 0]*C[0, 3]-C[1, 0]*C[1, 3]-C[2, 0]*C[2, 3]],
                    [C[0, 1], C[1, 1], C[2, 1], 
                    -C[0, 1]*C[0, 3]-C[1, 1]*C[1, 3]-C[2, 1]*C[2, 3]],
                    [C[0, 2], C[1, 2], C[2, 2], -C[0, 2] *
                        C[0, 3]-C[1, 2]*C[1, 3]-C[2, 2]*C[2, 3]],
                    [0, 0, 0, 1]])
    return CInv

def SO3Exp(axis: MutableDenseMatrix, angle: float | Expr) -> MutableDenseMatrix:
    """Compute exponential mapping for SO(3).

    Args:
        axis (sympy.Matrix): Rotation axis
        angle (double): Rotation angle

    Returns:
        sympy.Matrix: Rotation matrix
    """
    axis = Matrix(axis)
    xhat = Matrix([[0, -axis[2, 0], axis[1, 0]],
                    [axis[2, 0], 0, -axis[0, 0]],
                    [-axis[1, 0], axis[0, 0], 0]])
    R = Matrix(Identity(3)) + sin(angle) * xhat + (1-cos(angle))*(xhat*xhat)
    return R

def inertia_matrix(Ixx: float | Expr=0, Ixy: float | Expr=0, 
                   Ixz: float | Expr=0, Iyy: float | Expr=0, 
                   Iyz: float | Expr=0, Izz: float | Expr=0) -> MutableDenseMatrix:
    """Create 3 x 3 inertia matrix from independent inertia values.

    Args:
        Ixx (float or sympy.Expr): Inertia value I11. Defaults to 0.
        Ixy (float or sympy.Expr): Inertia value I12. Defaults to 0.
        Ixz (float or sympy.Expr): Inertia value I13. Defaults to 0.
        Iyy (float or sympy.Expr): Inertia value I22. Defaults to 0.
        Iyz (float or sympy.Expr): Inertia value I23. Defaults to 0.
        Izz (float or sympy.Expr): Inertia value I33. Defaults to 0.

    Returns:
        sympy.Matrix: Inertia matrix (3,3)
    """
    I = Matrix([[Ixx, Ixy, Ixz],
                [Ixy, Iyy, Iyz],
                [Ixz, Iyz, Izz]])
    return I

def transformation_matrix(r: MutableDenseMatrix=Matrix(Identity(3)), 
                         t: MutableDenseMatrix | list=zeros(3, 1)) -> MutableDenseMatrix:
    """Build transformation matrix from rotation and translation.

    Args:
        r (sympy.Matrix): 
            SO(3) Rotation matrix (3,3). 
            Defaults to sympy.Matrix(Identity(3))
        t (sympy.Matrix | list): 
            Translation vector (3,1). Defaults to sympy.zeros(3,1)

    Returns:
        sympy.Matrix: Transformation matrix (4,4)
    """
    if type(t) is list:
        t = Matrix(t)
    T = Matrix(Identity(4))
    T[:3,:3] = r
    T[:3,3] = t
    return T

def mass_matrix_mixed_data(m: float | Expr, Theta: MutableDenseMatrix, 
                        COM: MutableDenseMatrix, frame: str = "body",
                        R: MutableDenseMatrix = Matrix(Identity(3)),
                        ) -> MutableDenseMatrix:
    """Build mass-inertia matrix in SE(3) from mass, inertia and 
    center of mass information.

    Args:
        m (float | sympy.Expr): Mass.
        Theta (array_like): (3,3) Inertia tensor with respect to body-fixed
            or center of mass frame, depending on value of `frame`.
        COM (array_like): Center of mass (3,1).
        frame (str, optional): 
            reference frame of inertia tensor. Has to be "body" or "com".
            Defaults to "body". 
        R (array_like): (3,3) SO3 rotation matrix of the COM frame to
            the body frame. Only if frame == "com". Defaults to Identity.

    Returns:
        sympy.Matrix: Mass-inertia matrix (6,6).
    """
    assert frame in {"com", "body", "COM", "BODY"}
    if frame in {"com", "COM"}:
        R = Matrix(R)
        Theta = Matrix(Theta)
        Theta = R * Theta * R.T
        Theta -= Matrix([
            [-m * (COM[1]**2 + COM[2]**2), m * COM[0] * COM[1], m * COM[0] * COM[2]],
            [m * COM[0] * COM[1], -m * (COM[0]**2 + COM[2]**2), m * COM[1] * COM[2]],
            [m * COM[0] * COM[2], m * COM[1] * COM[2], -m * (COM[0]**2 + COM[1]**2)]
        ])
    M = Matrix([[Theta[0, 0], Theta[0, 1], Theta[0, 2], 0, (-COM[2])*m, COM[1]*m],
                [Theta[0, 1], Theta[1, 1], Theta[1, 2], COM[2]*m, 0, (-COM[0])*m],
                [Theta[0, 2], Theta[1, 2], Theta[2, 2], (-COM[1])*m, COM[0]*m, 0],
                [0, COM[2]*m, (-COM[1])*m, m, 0, 0],
                [(-COM[2])*m, 0, COM[0]*m, 0, m, 0],
                [COM[1]*m, (-COM[0])*m, 0, 0, 0, m]])
    return M

def mass_matrix_mixed_data_identification(m: float | Expr, Theta: MutableDenseMatrix, 
                        h: MutableDenseMatrix) -> MutableDenseMatrix:
    """Build mass-inertia matrix in SE(3) from mass, inertia and 
    first moment of mass, i.e. m*com_vector.

    Args:
        m (float): Mass.
        Theta (array_like): Inertia (3,3).
        h (array_like): First moment of mass, i.e. m*com_vector. (3,1).

    Returns:
        sympy.Matrix: Mass-inertia matrix (6,6).
    """
    M = Matrix([[Theta[0, 0], Theta[0, 1], Theta[0, 2], 0, 
                    (-h[2]), h[1]],
                [Theta[0, 1], Theta[1, 1], Theta[1, 2],
                    h[2], 0, (-h[0])],
                [Theta[0, 2], Theta[1, 2], Theta[2, 2],
                    (-h[1]), h[0], 0],
                [0, h[2], (-h[1]), m, 0, 0],
                [(-h[2]), 0, h[0], 0, m, 0],
                [h[1], (-h[0]), 0, 0, 0, m]])
    return M

def mass_matrix_to_parameter_vector(
    M: MutableDenseMatrix
    ) -> tuple[Expr,Expr,Expr,Expr,Expr,Expr,Expr,Expr,Expr,Expr]:
    """Get parameter vector for regressor matrix from mass-inertia matrix.
        M -> m, m*cx, m*cy, m*cz, Ixx, Ixy, Ixz, Iyy, Iyz, Izz
    
    Args:
        M (MutableDenseMatrix): (6,6) Mass-inertia matrix.
        
    Returns:
        tuple[Expr]: 
            (m, m*cx, m*cy, m*cz, Ixx, Ixy, Ixz, Iyy, Iyz, Izz).
    """
    m = M[3,3]
    I = M[:3,:3] 
    cxyz = Matrix([M[2,4], M[0,5], M[1,3]])
    return m, *cxyz, I[0,0], I[0,1], I[0,2], I[1,1], I[1,2], I[2,2]

def rpy_to_matrix(coords: list | MutableDenseMatrix) -> MutableDenseMatrix:
    """Convert roll-pitch-yaw coordinates to a 3x3 homogenous rotation matrix.

    The roll-pitch-yaw axes in a typical URDF are defined as a
    rotation of ``r`` radians around the x-axis followed by a rotation of
    ``p`` radians around the y-axis followed by a rotation of ``y`` radians
    around the z-axis. These are the Z1-Y2-X3 Tait-Bryan angles. See
    Wikipedia_ for more information.

    .. _Wikipedia: https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix

    Args:
        coords (MutableDenseMatrix | list): 
            The (3,) roll-pitch-yaw coordinates in order (x-rot, y-rot, z-rot).

    Returns:
        MutableDenseMatrix: 
            The corresponding homogenous 3x3 rotation matrix.        
    """
    c3 = cos(coords[0])
    c2 = cos(coords[1])
    c1 = cos(coords[2])
    s3 = sin(coords[0])
    s2 = sin(coords[1])
    s1 = sin(coords[2])
    return Matrix([
        [c1 * c2, (c1 * s2 * s3) - (c3 * s1), (s1 * s3) + (c1 * c3 * s2)],
        [c2 * s1, (c1 * c3) + (s1 * s2 * s3), (c3 * s1 * s2) - (c1 * s3)],
        [-s2, c2 * s3, c2 * c3]
    ])

def xyz_rpy_to_matrix(xyz_rpy: list | MutableDenseMatrix) -> MutableDenseMatrix:
    """Convert xyz_rpy coordinates to a 4x4 homogenous matrix.

    Args:
        xyz_rpy (MutableDenseMatrix | list): 
            The (6,) xyz_rpy vector.

    Returns:
        MutableDenseMatrix: 
            The (4,4) homogenous transform matrix.
    """
    matrix = Matrix(Identity(4))
    matrix[:3, 3] = xyz_rpy[:3]
    matrix[:3, :3] = rpy_to_matrix(xyz_rpy[3:])
    return matrix

def matrix_to_rpy(R: ArrayLike, solution: int=1) -> np.ndarray:
    """Convert a 3x3 transform matrix to roll-pitch-yaw coordinates.

    The roll-pitchRyaw axes in a typical URDF are defined as a
    rotation of ``r`` radians around the x-axis followed by a rotation of
    ``p`` radians around the y-axis followed by a rotation of ``y`` radians
    around the z-axis. These are the Z1-Y2-X3 Tait-Bryan angles. See
    Wikipedia_ for more information.

    .. _Wikipedia: https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix

    There are typically two possible roll-pitch-yaw coordinates that could have
    created a given rotation matrix. Specify ``solution=1`` for the first one
    and ``solution=2`` for the second one.

    Args:
        R (ArrayLike): 
            A 3x3 homogenous rotation matrix.
        solution (int, optional): 
            Either 1 or 2, indicating which solution to return. 
            Defaults to 1.

    Returns:
        ndarray:
            The (3,) roll-pitch-yaw coordinates in order (x-rot, y-rot, z-rot).
    """
    R = np.asanyarray(R, dtype=np.float64)
    r = 0.0
    p = 0.0
    y = 0.0

    if np.abs(R[2,0]) >= 1.0 - 1e-12:
        y = 0.0
        if R[2,0] < 0:
            p = np.pi / 2
            r = np.arctan2(R[0,1], R[0,2])
        else:
            p = -np.pi / 2
            r = np.arctan2(-R[0,1], -R[0,2])
    else:
        if solution == 1:
            p = -np.arcsin(R[2,0])
        else:
            p = np.pi + np.arcsin(R[2,0])
        r = np.arctan2(R[2,1] / np.cos(p), R[2,2] / np.cos(p))
        y = np.arctan2(R[1,0] / np.cos(p), R[0,0] / np.cos(p))

    return np.array([r, p, y], dtype=np.float64)


def matrix_to_xyz_rpy(matrix: np.ndarray) -> np.ndarray:
    """Convert a 4x4 homogenous matrix to xyzrpy coordinates.
    
    Args:
        matrix (ndarray): 
            The (4, 4) homogenous transform matrix.

    Returns:
        ndarray: The (6,) xyz_rpy vector.
    """
    xyz = matrix[:3,3]
    rpy = matrix_to_rpy(matrix[:3,:3])
    return np.hstack((xyz, rpy))

def quaternion_to_matrix(Q: list | MutableDenseMatrix) -> MutableDenseMatrix:
    """Convert a quaternion into SO(3) rotation matrix.

    Args:
        Q (list | sympy.MutableDenseMatrix): Quaternion in order [w,x,y,z].
        
    Returns:
        sympy.Matrix: (3,3) Rotation matrix.
    """
    # ensure symbolic values
    Q = Matrix(Q)  
    
    # Extract the values from Q
    q0 = Q[0]
    q1 = Q[1]
    q2 = Q[2]
    q3 = Q[3]

    # First row of the rotation matrix
    r00 = 2 * (q0 * q0 + q1 * q1) - 1
    r01 = 2 * (q1 * q2 - q0 * q3)
    r02 = 2 * (q1 * q3 + q0 * q2)

    # Second row of the rotation matrix
    r10 = 2 * (q1 * q2 + q0 * q3)
    r11 = 2 * (q0 * q0 + q2 * q2) - 1
    r12 = 2 * (q2 * q3 - q0 * q1)

    # Third row of the rotation matrix
    r20 = 2 * (q1 * q3 - q0 * q2)
    r21 = 2 * (q2 * q3 + q0 * q1)
    r22 = 2 * (q0 * q0 + q3 * q3) - 1

    # 3x3 rotation matrix
    rot_matrix = Matrix([[r00, r01, r02],
                         [r10, r11, r12],
                         [r20, r21, r22]])

    return rot_matrix


from skidy import SymbolicKinDyn, mass_matrix_mixed_data, inertia_matrix, transformation_matrix
from sympy import symbols, Matrix, Identity

skd = SymbolicKinDyn()
q1, q2, q3, q4 = symbols("q1 q2 q3 q4", real = True)
dq1, dq2, dq3, dq4 = symbols("dq1 dq2 dq3 dq4", real=True)
ddq1, ddq2, ddq3, ddq4 = symbols("ddq1 ddq2 ddq3 ddq4", real = True)

m1, c1x, c1y,c1z, I1, I1xx, I1xy, I1xz, I1yy, I1yz, I1zz = symbols("m1, c1x, c1y,c1z, I1, I1xx, I1xy, I1xz, I1yy, I1yz, I1zz", real=1, constant=1)
m2, c2x, c2y,c2z, I2, I2xx, I2xy, I2xz, I2yy, I2yz, I2zz = symbols("m2, c2x, c2y,c2z, I2, I2xx, I2xy, I2xz, I2yy, I2yz, I2zz", real=1, constant=1)
m3, c3x, c3y,c3z, I3, I3xx, I3xy, I3xz, I3yy, I3yz, I3zz = symbols("m3, c3x, c3y,c3z, I3, I3xx, I3xy, I3xz, I3yy, I3yz, I3zz", real=1, constant=1)
m4, c4x, c4y,c4z, I4, I4xx, I4xy, I4xz, I4yy, I4yz, I4zz = symbols("m4, c4x, c4y,c4z, I4, I4xx, I4xy, I4xz, I4yy, I4yz, I4zz", real=1, constant=1)

pi, g = symbols("pi g", real=1, constant=1)

base_offset, thigh_v_offset, thigh_h_offset, thigh, shank = symbols("base_offset thigh_v_offset thigh_h_offset thigh shank")

skd.gravity_vector = Matrix([0, 0, g])

dh_table = Matrix([[0,0,0,0,base_offset],
                   [0, pi/2, 0, 0, thigh_v_offset],
                   [0,-pi/2, 0, -pi/2, thigh_h_offset],
                   [0, 0, thigh, pi, 0]])

skd.dh_to_screw_coord(dh_table)
re = Matrix([shank,0,0])
skd.ee = transformation_matrix(t=re)

skd.Mb = []
skd.Mb.append(mass_matrix_mixed_data(m1, inertia_matrix(I1xx, I1xy, I1xz, I1yy, I1yz, I1zz),Matrix([c1x,c1y,c1z])))
skd.Mb.append(mass_matrix_mixed_data(m2, inertia_matrix(I2xx, I2xy, I2xz, I2yy, I2yz, I2zz),Matrix([c2x,c2y,c2z])))
skd.Mb.append(mass_matrix_mixed_data(m3, inertia_matrix(I3xx, I3xy, I3xz, I3yy, I3yz, I3zz),Matrix([c3x,c3y,c3z])))
skd.Mb.append(mass_matrix_mixed_data(m4, inertia_matrix(I4xx, I4xy, I4xz, I4yy, I4yz, I4zz),Matrix([c4x,c4y,c4z])))

q = Matrix([q1,q2,q3,q4])
qd = Matrix([dq1,dq2,dq3,dq4])
q2d = Matrix([ddq1,ddq2,ddq3,ddq4])

T = skd.closed_form_kinematics_body_fixed(q,qd,q2d,simplify=0)

Q = skd.closed_form_inv_dyn_body_fixed(q,qd,q2d,simplify=0)

# s.generate_code(python=True)



from sympy import symbols, Matrix, Identity
from skidy import SymbolicKinDyn as skd
from skidy import transformation_matrix, mass_matrix_mixed_data

# Declaration of symbolic variables
q1, q2, q3 = symbols("q1 q2 q3")
dq1, dq2, dq3 = symbols("dq1 dq2 dq3")
ddq1, ddq2, ddq3 = symbols("ddq1 ddq2 ddq3")

m1, m2, m3, I1, I2, I3 = symbols("m1, m2, m3, I1, I2, I3",real=True)
Lc1, Lc2, Lc3, g = symbols("Lc1, Lc2, Lc3, g", real=True)
L1, L2 = symbols("L1, L2", real=True)

gravity_vector = Matrix([0,g,0])

# Graph description
parent = [0,
          0,
          2]
child = [[],
         [3],
         []]
support = [[1],
           [2],
           [2,3]]

# joint screw coordinates
Y = []

# For revolute joint
e1 = Matrix([0,0,1])
y1 = Matrix([0,0,0])
Y.append(Matrix([e1,y1.cross(e1)]))

# For revolute joint
e2 = Matrix([0,0,1])
y2 = Matrix([L1,0,0])
Y.append(Matrix([e2,y2.cross(e2)]))

# For prismatic joint
e3 = Matrix([1,0,0])
Y.append(Matrix([Matrix([0,0,0]),e3]))

# Reference configurations of body (i.e. w.r.t. inertia frame)
A = []

A.append(Matrix(Identity(4)))

r2 = Matrix([L1,0,0])
A.append(transformation_matrix(t=r2))

r3 = Matrix([L1,0,0])
A.append(transformation_matrix(t=r3))

# End-effector configuration wrt last link body fixed frame in the chain
re = Matrix([L2, 0, 0])
ee = transformation_matrix(t=re)

# Mass Inertia Parameters
Mb = []
Mb.append(mass_matrix_mixed_data(m1,I1*Identity(3),Matrix([Lc1, 0, 0])))
Mb.append(mass_matrix_mixed_data(m2,I2*Identity(3),Matrix([Lc2, 0, 0])))
Mb.append(mass_matrix_mixed_data(m3,I3*Identity(3),Matrix([Lc3, 0, 0])))

# EOM for the subsystem II i.e. RP Leg
q = Matrix([q1,q2,q3])
qd = Matrix([dq1,dq2,dq3])
q2d = Matrix([ddq1,ddq2,ddq3])

s = skd(gravity_vector=gravity_vector,ee=ee,Mb=Mb,parent=parent,support=support,child=child,A=A,Y=Y)
print(s.closed_form_kinematics_body_fixed(q,qd,q2d, simplify=True))
print(s.closed_form_inv_dyn_body_fixed(q,qd,q2d,simplify=True))
s.generate_code(python=True, C=True, Matlab=True,
                use_global_vars=True, name="lambdaMechanism", project="LambdaMechanism")


from skidy import SymbolicKinDyn
import os

from sympy import symbols, Matrix, Identity

q0, q1, q2 = symbols("q0 q1 q2")
dq0, dq1, dq2 = symbols("dq0 dq1 dq2")
ddq0, ddq1, ddq2 = symbols("ddq0 ddq1 ddq2")
g, L1,L2, Ixx1, Ixy1, Ixz1, Iyy1, Iyz1, Izz1, Ixx2, Ixy2, Ixz2, Iyy2, Iyz2, Izz2, m0, m1, m2, cx1,cy1,cz1, cx2, cy2, cz2 = symbols("g L1 L2 Ixx1 Ixy1 Ixz1 Iyy1 Iyz1 Izz1 Ixx2 Ixy2 Ixz2 Iyy2 Iyz2 Izz2 m0 m1 m2 cx1 cy2 cz1 cx2 cy2 cz2", constant=True, real = True)
    

s = SymbolicKinDyn()


s.gravity_vector = Matrix([0,0,g])

# end effector
re = Matrix([0, 0, L2])
s.ee = Matrix(Identity(3)).row_join(re).col_join(Matrix([0, 0, 0, 1]).T)

s.load_from_urdf(os.path.join(os.path.dirname(__file__),"urdf/hopping_leg.urdf"),symbolic=1, simplify_numbers=1, cse=1)


q = Matrix([q0,q1, q2])
qd = Matrix([dq0, dq1, dq2])
q2d = Matrix([ddq0, ddq1, ddq2])

# Kinematics
F = s.closed_form_kinematics_body_fixed(q, qd, q2d, simplify=0,cse=1)
# Q = s.closed_form_inv_dyn_body_fixed(q,qd,q2d, simplify=0)

s.generate_code(python=1,C=0,Matlab=0,name="HoppingLeg")



from __future__ import annotations

import yaml
import json
from typing import Iterable, Optional
from skidy.kinematics_generator import SymbolicKinDyn
from skidy.matrices import (joint_screw, 
                            symbolic_inertia_matrix, mass_matrix_mixed_data, 
                            SO3Exp, transformation_matrix, inertia_matrix,
                            quaternion_to_matrix, xyz_rpy_to_matrix,
                            rpy_to_matrix)
from sympy import Matrix, Identity, parse_expr, Expr, zeros
import regex

def robot_from_yaml(path: str) -> SymbolicKinDyn:
    """Parse yaml robot description and return SymbolicKinDyn Object.

    Args:
        path (str): Path to yaml file.

    Returns:
        skidy.SymbolicKinDyn object.
    """
    with open(path, "r") as stream:
        y = yaml.safe_load(stream)
    return dict_parser(y)

def robot_from_json(path: str) -> SymbolicKinDyn:
    """Parse json robot description and return SymbolicKinDyn Object.

    Args:
        path (str): Path to json file.

    Returns:
        skidy.SymbolicKinDyn object.
    """
    with open(path, "r") as stream:
        y = json.load(stream)
    return dict_parser(y)

def parse_hierarchical_expr(x: list | dict | str, 
                            include_keys:Iterable={}, 
                            exclude_keys:Iterable={}) -> list | dict | Expr:
    """Convert strings in data structure (list or dict) to sympy 
    expressions.

    Args:
        x (list | dict | str): 
            Hierarchical structure which might contain symbolic 
            expression as string. 
        include_keys (Iterable): 
            Only convert strings with listed dict_keys. 
            Defaults to {}.
        exclude_keys (Iterable): 
            Don't convert strings with listed dict_keys. 
            Defaults to {}.
            
    Returns:
        list | dict | sympy.Expr: 
            Same data structure with strings converted to sympy.Expr. 
    """
    if type(x) in {dict, list}:
        for i in x if type(x) is dict else range(len(x)):
            if type(x[i]) in {list,dict}:
                x[i] = parse_hierarchical_expr(x[i],include_keys,exclude_keys)
            elif type(x[i]) is str:
                if (type(x) is list 
                    or (include_keys and i in include_keys) 
                    or (not include_keys and i not in exclude_keys)):
                    x[i] = parse_expr(x[i])
    elif type(x) is str:
        x = parse_expr(x)
    return x

def dict_parser(d: dict) -> SymbolicKinDyn:
    """Parse dict to SymbolicKinDyn object.

    Args:
        d (dict): Dictionary containing robot description.

    Raises:
        KeyError: Entry not found.
        ValueError: Unexpected entry.

    Returns:
        skidy.SymbolicKinDyn object.
    """
    d = parse_hierarchical_expr(d, include_keys={"mass", "index", 
                                                 "Ixx", "Ixy", "Ixz", 
                                                 "Iyy", "Ixz", "Izz", 
                                                 "q", "qd", "q2d", "q3d", "q4d",
                                                 "WEE", "WDEE", "W2DEE",
                                                 "inertia"})
    
    config_representation = d["representation"] if "representation" in d else "spatial"
    
    parent = d["parent"] if "parent" in d else []
    child = d["child"] if "child" in d else []
    support = d["support"] if "support" in d else []
    ee_parent = d["ee_parent"] if "ee_parent" in d else None
    
    try:
        gravity = d["gravity"] if "gravity" in d else d["gravity_vector"]
    except KeyError:
        gravity = None
    
    assert("joint_screw_coord" in d)
    joint_screw_coord = []
    for js in d["joint_screw_coord"]:
        if type(js) is list and len(js) == 6:
            joint_screw_coord.append(Matrix(js))
        elif type(js) is dict:
            if js["type"] == "prismatic":
                joint_screw_coord.append(
                    joint_screw(js["axis"],revolute=0))
            elif js["type"] == "revolute":
                joint_screw_coord.append(
                    joint_screw(
                        js["axis"],
                        js["vec"] if "vec" in js else [0,0,0],
                        revolute=1))
            else:
                raise ValueError(f"joint type {js['type']} not supported.")
        else:
            raise ValueError("joint screw coordinates corrupted.") 

    assert("body_ref_config" in d)
    
    body_ref_config = []
    for br in d["body_ref_config"]:
        if type(br) is list:
            body_ref_config.append(Matrix(br))
        elif type(br) is dict:
            if "xyzrpy" in br:
                body_ref_config.append(xyz_rpy_to_matrix(br["xyzrpy"]))                
            else:
                t = br["translation"] if "translation" in br else [0,0,0]
                if "rotation" in br:
                    if type(br["rotation"]) is dict:
                        if "Q" in br["rotation"]:
                            r = quaternion_to_matrix(br["rotation"]["Q"])
                        elif "rpy" in br["rotation"]:
                            r = rpy_to_matrix(br["rotation"]["rpy"])
                        elif "axis" in br["rotation"]:
                            axis = br["rotation"]["axis"] if "axis" in br["rotation"] else [0,0,1]
                            angle = br["rotation"]["angle"] if "angle" in br["rotation"] else 0
                            r = SO3Exp(Matrix(axis),angle)
                    else:
                        r = Matrix(br["rotation"])
                else:
                    r = Matrix(Identity(3))
                body_ref_config.append(transformation_matrix(r,t))
    
    assert("ee" in d)
    
    # transformation matrix directly in ee
    if type(d["ee"]) is list and type(d["ee"][0]) is list and type(d["ee"][0][0]) is not list:
        ee = Matrix(d["ee"])
    # dict in directly in ee
    elif type(d["ee"]) is dict:
        if "xyzrpy" in d["ee"]:
            ee = xyz_rpy_to_matrix(d["ee"]["xyzrpy"])
        else:
            t = d["ee"]["translation"] if "translation" in d["ee"] else [0,0,0]
            if "rotation" in d["ee"]:
                if type(d["ee"]["rotation"]) is dict:
                    if "Q" in d["ee"]["rotation"]:
                        r = quaternion_to_matrix(d["ee"]["rotation"]["Q"])
                    elif "rpy" in d["ee"]["rotation"]:
                        r = rpy_to_matrix(d["ee"]["rotation"]["rpy"])
                    elif "axis" in d["ee"]["rotation"]:
                        axis = d["ee"]["rotation"]["axis"] if "axis" in d["ee"]["rotation"] else [0,0,1]
                        angle = d["ee"]["rotation"]["angle"] if "angle" in d["ee"]["rotation"] else 0
                        r = SO3Exp(Matrix(axis),angle)
                else:
                    r = Matrix(d["ee"]["rotation"])
            else:
                r = Matrix(Identity(3))
            ee = transformation_matrix(r,t)
    # list of ee
    elif type(d["ee"]) is list:
        ee = []
        for eei in d["ee"]:
            if type(eei) is list:
                ee.append(Matrix(eei))
            elif type(eei) is dict:
                if "xyzrpy" in eei:
                    ee.append(xyz_rpy_to_matrix(eei["xyzrpy"]))                
                else:
                    t = eei["translation"] if "translation" in eei else [0,0,0]
                    if "rotation" in eei:
                        if type(eei["rotation"]) is dict:
                            if "Q" in eei["rotation"]:
                                r = quaternion_to_matrix(eei["rotation"]["Q"])
                            elif "rpy" in eei["rotation"]:
                                r = rpy_to_matrix(eei["rotation"]["rpy"])
                            elif "axis" in eei["rotation"]:
                                axis = eei["rotation"]["axis"] if "axis" in eei["rotation"] else [0,0,1]
                                angle = eei["rotation"]["angle"] if "angle" in eei["rotation"] else 0
                                r = SO3Exp(Matrix(axis),angle)
                        else:
                            r = Matrix(eei["rotation"])
                    else:
                        r = Matrix(Identity(3))
                    ee.append(transformation_matrix(r,t))        
    else:
        raise ValueError(f"ee {d['ee']} cannot be processed.")
    
    
    Mb = []
    inertia_frame = "body" # default
    
    if "mass_inertia" in d:
        for mb in d["mass_inertia"]:
            if type(mb) is list:
                Mb.append(Matrix(mb))
            elif type(mb) is dict:
                if "mass" in mb and "inertia" in mb and "com" in mb:
                    if "frame" in mb:
                        if mb["frame"] in {"com", "body"}:
                            inertia_frame = mb["frame"]
                        else: 
                            raise ValueError(f"Unknown frame '{mb['inertia']}'")
                    # get rotation of inertia
                    if inertia_frame == "com" and "rotation" in mb:
                        if type(mb["rotation"]) is dict:
                            if "Q" in mb["rotation"]:
                                r = quaternion_to_matrix(mb["rotation"]["Q"])
                            elif "rpy" in mb["rotation"]:
                                r = rpy_to_matrix(mb["rotation"]["rpy"])
                            elif "axis" in mb["rotation"]:
                                axis = mb["rotation"]["axis"] if "axis" in mb["rotation"] else [0,0,1]
                                angle = mb["rotation"]["angle"] if "angle" in mb["rotation"] else 0
                                r = SO3Exp(Matrix(axis),angle)
                        else:
                            r = Matrix(mb["rotation"])
                    else:
                        r = Matrix(Identity(3))    
                        
                    if type(mb["inertia"]) is list:
                        if len(mb["inertia"]) == 6:
                            Mb.append(
                                mass_matrix_mixed_data(
                                    mb["mass"],
                                    inertia_matrix(*mb["inertia"]),
                                    Matrix(mb["com"]),
                                    inertia_frame,
                                    r
                                )
                            )
                        else:    
                            Mb.append(
                                mass_matrix_mixed_data(
                                    mb["mass"],
                                    Matrix(mb["inertia"]),
                                    Matrix(mb["com"]),
                                    inertia_frame,
                                    r
                                )
                            )
                    elif type(mb["inertia"]) is dict and "index" in mb["inertia"]:
                        Mb.append(
                            mass_matrix_mixed_data(
                                mb["mass"], 
                                symbolic_inertia_matrix(mb["inertia"]["index"],
                                                    mb["inertia"]["pointmass"]),
                                Matrix(mb["com"]),
                                inertia_frame,
                                r
                            )
                        )
                    elif type(mb["inertia"]) is dict:
                        Mb.append(
                            mass_matrix_mixed_data(
                                mb["mass"], 
                                inertia_matrix(**mb["inertia"]),
                                Matrix(mb["com"]),
                                inertia_frame,
                                r
                            )
                        )
                    else:
                        Mb.append(
                            mass_matrix_mixed_data(
                                mb["mass"], 
                                mb["inertia"]*Identity(3),
                                Matrix(mb["com"]),
                                inertia_frame,
                                r
                            )
                        )
                        # raise ValueError(f"Inertia {mb['inertia']} not supported.")
                else:
                    raise ValueError("Unable to process mass_inertia.")
            else:
                raise ValueError("Unable to process mass_inertia.")
    
    assignments = {}
    if "assignments" in d:
        if type(d["assignments"]) is dict:
            assignments = {parse_expr(i): d["assignments"][i] for i in d["assignments"] }
    
    q = Matrix(d["q"]) if "q" in d else None        
    qd = Matrix(d["qd"]) if "qd" in d else None        
    q2d = Matrix(d["q2d"]) if "q2d" in d else None        
    q3d = Matrix(d["q3d"]) if "q3d" in d else None        
    q4d = Matrix(d["q4d"]) if "q4d" in d else None
    if "WEE" in d:
        assert(type(d["WEE"] is list))
        if type(d["WEE"][0]) is list:
            WEE = [Matrix(w) for w in d["WEE"]]
        else:
            WEE = Matrix(d["WEE"])
    else:
        WEE = zeros(6,1)        
    if "WDEE" in d:
        assert(type(d["WDEE"] is list))
        if type(d["WDEE"][0]) is list:
            WDEE = [Matrix(w) for w in d["WDEE"]]
        else:
            WDEE = Matrix(d["WDEE"])
    else:
        WDEE = zeros(6,1)        
    if "W2DEE" in d:
        assert(type(d["W2DEE"] is list))
        if type(d["W2DEE"][0]) is list:
            W2DEE = [Matrix(w) for w in d["W2DEE"]]
        else:
            W2DEE = Matrix(d["W2DEE"])
    else:
        W2DEE = zeros(6,1)        
    
    skd = SymbolicKinDyn(gravity_vector=gravity,ee=ee,
                         body_ref_config=body_ref_config,
                         joint_screw_coord=joint_screw_coord,
                         config_representation=config_representation, 
                         Mb=Mb, parent=parent, support=support, 
                         child=child, ee_parent=ee_parent,
                         q=q,qd=qd,q2d=q2d,q3d=q3d,q4d=q4d,
                         WEE=WEE, WDEE=WDEE, W2DEE=W2DEE, 
                         assignments=assignments)
    return skd

def generate_template_yaml(path: str="edit_me.yaml", structure: Optional[str]=None, 
                        dof: int=0, tree: bool=True, **kwargs) -> None:
    """Generate template yaml file to modify for own robot.

    Args:
        path (str, optional): Path where to save generated yaml file.
            Defaults to 'edit_me.yaml' 
        structure (str, optional): string containing only 'r' and 'p' of 
            joint order. Use "prr" for a robot which has 1 prismatic joint
            followed by 2 revolute joints. Defaults to None.
        dof (int, optional): Degrees of freedom. Is usually calculated 
            by length of 'structure'. Defaults to 0.
        tree (bool, optional): Generate parent, child and support array 
            in yaml file. Defaults to True.

    Raises:
        ValueError: Unexpected letter in 'structure'.
    """
    if type(structure) is str:
        if not dof:
           dof = len(structure)
        else:
            # fill structure with r until length matches dof
            while len(structure) < dof:
                structure += "r"
            
    elif dof:
        structure = "r"*dof
    y = ["---"]
    if tree:
        y.append("parent:")
        for i in range(dof):
            y.append(f"  - {i}")
        y.append("")
     
        y.append("child:")
        for i in range(dof):
            y.append(f"  - [{i+1 if i < dof-1 else ''}]")
        y.append("")
     
        y.append("support:")
        for i in range(dof):
            y.append(f"  - [{','.join([str(j) for j in range(1,i+2)])}]")
        y.append("")
        
        y.append(f"ee_parent: {dof}")
        y.append("")
        
    y.append("gravity: [0,0,-g]")
    y.append("")
    
    y.append("representation: spatial")
    y.append("")
    
    y.append("joint_screw_coord:")
    for i in range(dof):
        if structure:
            if structure[i] in "rR":
                y.append(f"  - type: revolute")
                y.append(f"    axis: [0,0,1]")
                y.append(f"    vec: [0,0,0]")
                y.append("")
            elif structure[i] in "pP":
                y.append(f"  - type: prismatic")
                y.append(f"    axis: [0,0,1]")
                y.append("")
            else:
                raise ValueError(f"structure {structure[i]} not supported.")
        else:
            y.append("  - []")
            y.append("")
    
    y.append("body_ref_config:")
    for i in range(dof):
        y.append("  - rotation:")
        y.append("      axis: [0,0,1]")
        y.append("      angle: 0")
        y.append("    translation: [0,0,0]")
        y.append("")
    
    y.append("ee:")
    y.append("  - rotation:")
    y.append("      axis: [0,0,1]")
    y.append("      angle: 0")
    y.append("    translation: [0,0,0]")
    y.append("")
    
    y.append("mass_inertia:")
    for i in range(dof):
        y.append(f"  - mass: m{i+1}")
        y.append( "    frame: body")
        y.append( "    inertia:")
        y.append(f"      Ixx: Ixx{i+1}")
        y.append(f"      Ixy: Ixy{i+1}")
        y.append(f"      Ixz: Ixz{i+1}")
        y.append(f"      Iyy: Iyy{i+1}")
        y.append(f"      Iyz: Iyz{i+1}")
        y.append(f"      Izz: Izz{i+1}")
        y.append( "    com: [0,0,0]")
        y.append("")
    
    y.append(f"q: [{','.join(['q'+str(i+1) for i in range(dof)])}]")
    y.append(f"qd: [{','.join(['dq'+str(i+1) for i in range(dof)])}]")
    y.append(f"q2d: [{','.join(['ddq'+str(i+1) for i in range(dof)])}]")
    y.append("WEE: [0,0,0,0,0,0]")
    y.append("")
    y.append("assignments:")
    y.append("  g: 9.81")
    y.append("")
    if "return_dict" in kwargs and kwargs["return_dict"]:
        return yaml.load("\n".join(y))
        
    
    # check if path ends with .yaml
    if not bool(regex.search("(.yaml|.YAML|.yml|.YML)\Z", path)):
        path += ".yaml"
        
    with open(path, "w+") as f:
        f.write("\n".join(y))
    
    
def generate_template_json(path: str="edit_me.json", structure: Optional[str]=None, 
                        dof: int=0, tree: bool=True) -> None:
    """Generate template json file to modify for own robot.

    Args:
        path (str, optional): Path where to save generated json file.
            Defaults to 'edit_me.json' 
        structure (str, optional): string containing only 'r' and 'p' of 
            joint order. Use "prr" for a robot which has 1 prismatic joint
            followed by 2 revolute joints. Defaults to None.
        dof (int, optional): Degrees of freedom. Is usually calculated 
            by length of 'structure'. Defaults to 0.
        tree (bool, optional): Generate parent, child and support array 
            in yaml file. Defaults to True.

    Raises:
        ValueError: Unexpected letter in 'structure'.
    """
    d = generate_template_yaml("",structure,dof,tree,return_dict=True)
    s= json.dumps(d,indent=2)
    # check if path ends with .json
    if not bool(regex.search("(.JSON|.json)\Z", path)):
        path += ".json"
    
    with open(path, "w+") as f:
        f.write(s)


def generate_template_python(path: str="edit_me.py", 
                             structure: Optional[str]=None, dof: int=0, 
                             tree: bool=True, urdf: bool=False) -> None:
    """Generate template python file to modify for own robot.

    Args:
        path (str, optional): Path where to save generated python file.
            Defaults to 'edit_me.py' 
        structure (str, optional): string containing only 'r' and 'p' of 
            joint order. Use "prr" for a robot which has 1 prismatic joint
            followed by 2 revolute joints. Defaults to None.
        dof (int, optional): Degrees of freedom. Is usually calculated 
            by length of 'structure'. Defaults to 0.
        tree (bool, optional): Generate parent, child and support array 
            in yaml file. Defaults to True.

    Raises:
        ValueError: Unexpected letter in 'structure'.
    """
    if type(structure) is str:
        if not dof:
           dof = len(structure)
        else:
            # fill structure with r until length matches dof
            while len(structure) < dof:
                structure += "r"
    elif dof:
        structure = "r"*dof
    
    p = ["from skidy import (SymbolicKinDyn,"]
    p.append("                   transformation_matrix,")
    if not urdf:
        p.append("                   mass_matrix_mixed_data,")
        p.append("                   joint_screw,")
        p.append("                   inertia_matrix,")
    p.append("                   SO3Exp,")
    p.append("                   generalized_vectors)")
    p.append("from skidy.symbols import g, pi")
    p.append("import sympy")
    p.append("")
    p.append("# Define symbols: (Hint: you can import the most common use symbols from skidy.symbols instead)")
    if not urdf:
        p.append(f"{'m'+ ', m'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'m'+ ' m'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
        p.append(f"{'Ixx'+ ', Ixx'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'Ixx'+ ' Ixx'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
        p.append(f"{'Ixy'+ ', Ixy'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'Ixy'+ ' Ixy'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
        p.append(f"{'Ixz'+ ', Ixz'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'Ixz'+ ' Ixz'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
        p.append(f"{'Iyy'+ ', Iyy'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'Iyy'+ ' Iyy'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
        p.append(f"{'Iyz'+ ', Iyz'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'Iyz'+ ' Iyz'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
        p.append(f"{'Izz'+ ', Izz'.join(str(i) for i in range(1, dof+1))} = sympy.symbols('{'Izz'+ ' Izz'.join(str(i) for i in range(1, dof+1))}', real=True, const=True)")
    else:
        p.append("lee = sympy.symbols('lee', real=True, const=True)")
    p.append("") # m1, m2,
    if tree and not urdf:
        p.append("# Define connectivity graph")
        p.append(f"parent = [0{',' if dof > 1 else ']'}")
        for i in range(1, dof):
            p.append(f"          {i}{',' if dof > i+1 else ']'}")
        p.append("")
     
        p.append(f"child = [[{1 if dof > 1 else ''}]{',' if dof > 1 else ']'}")
        for i in range(1, dof):
            p.append(f"         [{i+1 if dof > i+1 else ''}]{',' if dof > i+1 else ']'}")
        p.append("")
     
        p.append(f"support = [[1]{',' if dof > 1 else ']'}")
        for i in range(1,dof):
            p.append(f"           [{','.join([str(j) for j in range(1,i+2)])}]{',' if dof > i+1 else ']'}")
        p.append("")
        
        p.append(f"ee_parent = {dof}")
        p.append("")
        
    if urdf:
        p.append("urdfpath = '/path/to/urdf' # TODO: change me!")
        p.append("")
    
    p.append("# gravity vector")
    p.append("gravity = sympy.Matrix([0,0,-g])")
    p.append("")
    
    if not urdf:
        p.append("# representation of joint screw coordinates and body reference configurations")
        p.append("representation = 'spatial' # alternative: 'body_fixed'")
        p.append("")
        
        p.append("# joint screw coordinates (6x1 sympy.Matrix per joint)")
        p.append("joint_screw_coord = []")
        for i in range(dof):
            if structure:
                if structure[i] in "rR":
                    p.append('joint_screw_coord.append(joint_screw(axis=[0,0,1], vec=[0,0,0], revolute=True))')
                elif structure[i] in "pP":
                    p.append('joint_screw_coord.append(joint_screw(axis=[0,0,1], revolute=False))')
                else:
                    raise ValueError(f"structure {structure[i]} not supported.")
            else:
                p.append("  - []")
                p.append("")
        p.append("")
        
        p.append("# body reference configurations (4x4 SE3 Pose (sympy.Matrix) per link)")
        p.append("body_ref_config = []")
        for i in range(dof):
            p.append("body_ref_config.append(transformation_matrix(r=SO3Exp(axis=[0,0,1],angle=0),t=[0,0,0]))")
        p.append("")
    
    p.append("# end-effector configuration w.r.t. last link body fixed frame in the chain (4x4 SE3 Pose (sympy.Matrix))")
    p.append("ee = transformation_matrix(r=SO3Exp(axis=[0,0,1],angle=0),t=[0,0,0])")
    p.append("")
    
    if not urdf:
        p.append("# mass_inertia parameters (6x6 sympy.Matrix per link)")
        p.append("Mb = []")
        for i in range(dof):
            p.append(f"Mb.append(mass_matrix_mixed_data(m{i+1}, inertia_matrix(Ixx{i+1},Ixy{i+1},Ixz{i+1},Iyy{i+1},Iyz{i+1},Izz{i+1}), sympy.Matrix([0,0,0]), frame='body'))")
        p.append("")
        
        p.append(f"q, qd, q2d = generalized_vectors(len(body_ref_config), startindex=1)")
        p.append("WEE = sympy.zeros(6,1)")
        p.append("")
    
    p.append("skd = SymbolicKinDyn(gravity_vector=gravity,")
    p.append("                     ee=ee,")
    if not urdf: 
        p.append("                     body_ref_config=body_ref_config,")
        p.append("                     joint_screw_coord=joint_screw_coord,")
        p.append("                     config_representation=representation,")
        p.append("                     Mb=Mb,")
        if tree:
            p.append("                     parent=parent,")
            p.append("                     child=child,")
            p.append("                     support=support,")
            p.append("                     ee_parent=ee_parent,")
    p.append("                     )")
    p.append("")
    if urdf:
        p.append("skd.load_from_urdf(path = urdfpath,")
        p.append("                   symbolic=False, # symbolify equations? (eg. use Ixx instead of numeric value)")
        p.append("                   cse=False, # use common subexpression elimination? ")
        p.append("                   simplify_numbers=True, # round numbers if close to common fractions like 1/2 etc and replace eg 3.1416 by pi?")
        p.append("                   tolerance=0.0001, # tolerance for simplify numbers")
        p.append("                   max_denominator=8, # define max denominator for simplify numbers to avoid simplification to something like 13/153")
        p.append("                   )")
        p.append("")
    
        p.append(f"q, qd, q2d = generalized_vectors(len(skd.body_ref_config), startindex=1)")
        p.append("WEE = sympy.zeros(6,1)")
        p.append("")
    
    p.append("# run Calculations")
    p.append("skd.closed_form_kinematics_body_fixed(q, qd, q2d, simplify=True)")
    p.append("skd.closed_form_inv_dyn_body_fixed(q, qd, q2d, WEE=WEE, simplify=True)")
    p.append("")
    
    p.append("# Generate Code")
    p.append('skd.generate_code(python=True, cpp=False, C=False, Matlab=False, latex=False,')
    p.append('                 folder="./generated_code", use_global_vars=True,')
    p.append('                 name="plant", project="Project")')
    p.append("")
    
    
    # check if path ends with .yaml
    if not bool(regex.search("(.py)\Z", path)):
        path += ".py"
        
    with open(path, "w+") as f:
        f.write("\n".join(p))


def robot_from_urdf(path: str, symbolic: bool=False, cse: bool=False, 
                    simplify_numbers: bool=True, tolerance: float=0.0001, 
                    max_denominator: int=9) -> SymbolicKinDyn:
    """Load robot from urdf.

        Args:
            path (str): path to URDF.
            symbolic (bool, optional): 
                generate symbols for numeric values. 
                Defaults to False.
            cse (bool, optional): 
                use common subexpression elimination. Defaults to False.
            simplify_numbers (bool, optional): 
                Use eg. pi/2 instead of 1.5708. Defaults to True.
            tolerance (float, optional): 
                tolerance for simplify_numbers. Defaults to 0.0001.
            max_denominator (int, optional): 
                Maximum denominator to use for simplify numbers to avoid
                values like 13/153. Use 0 to deactivate. Defaults to 9.

        Raises:
            NotImplementedError: supports only the joint types 
                "revolute", "continuous" and "prismatic".
        
        Returns:
            skidy.SymbolicKinDyn object.
    """            
    skd = SymbolicKinDyn()
    skd.load_from_urdf(path=path, symbolic=symbolic, cse=cse, 
                       simplify_numbers=simplify_numbers, tolerance=tolerance, 
                       max_denominator=max_denominator)
    return skd

def skd_to_yaml(skd: SymbolicKinDyn, path: str="robot.yaml", **kwargs) -> None | dict:
    """Generate YAML file from SymbolicKinDyn object.

    Args:
        skd (SymbolicKinDyn): SymbolicKinDyn with set parameters.
        path (str, optional): path where to save .yaml file. 
            Defaults to "robot.yaml".
    """
    y = ["---"]
    dof = skd.n or len(skd.body_ref_config)
    if skd.parent is not None:
        y.append("parent:")
        for i in skd.parent:
            y.append(f"  - {i}")
        y.append("")

    if skd.child is not None:
        y.append("child:")
        for i in skd.child:
            y.append(f"  - {i}")
        y.append("")
    
    if skd.support is not None:
        y.append("support:")
        for i in skd.support:
            y.append(f"  - {i}")
        y.append("")
        
    if skd.ee_parent is not None:
        y.append(f"ee_parent: {skd.ee_parent}")
        y.append("")
        
    if skd.gravity_vector is not None:
        y.append(f"gravity: {[str(i) for i in skd.gravity_vector]}")
        y.append("")
    
    y.append(f"representation: {skd.config_representation}")
    y.append("")
    
    y.append("joint_screw_coord:")
    for i in skd.joint_screw_coord:
        y.append(f"  - {[str(j) for j in i]}")
    y.append("")
    
    y.append("body_ref_config:")
    for i in skd.body_ref_config:
        y.append("  - rotation:")
        y.append(f"      [{[str(j) for j in i[0,:3]]},")
        y.append(f"       {[str(j) for j in i[1,:3]]},")
        y.append(f"       {[str(j) for j in i[2,:3]]}]")
        y.append(f"    translation: {[str(j) for j in i[:3,3]]}")
        y.append("")
    
    y.append("ee:")
    if skd.ee is not None: 
        for i in skd.ee if type(skd.ee) is list else [skd.ee]:
            y.append("  - rotation:")
            y.append(f"      [{[str(j) for j in i[0,:3]]},")
            y.append(f"       {[str(j) for j in i[1,:3]]},")
            y.append(f"       {[str(j) for j in i[2,:3]]}]")
            y.append(f"    translation: {[str(j) for j in i[:3,3]]}")
            y.append("")
    else:
        y.append("  - rotation:")
        y.append("      axis: [0,0,1]")
        y.append("      angle: 0")
        y.append("    translation: [0,0,0]")
        y.append("")
    
    if skd.Mb is not None:
        y.append("mass_inertia:")
        for i in skd.Mb:
            if False:
                y.append(f"  - [{[str(j) for j in i[0,:]]},")
                y.append(f"     {[str(j) for j in i[1,:]]},")
                y.append(f"     {[str(j) for j in i[2,:]]},")
                y.append(f"     {[str(j) for j in i[3,:]]},")
                y.append(f"     {[str(j) for j in i[4,:]]},")
                y.append(f"     {[str(j) for j in i[5,:]]}]")
                y.append("")
            else:
                y.append(f"  - mass: {str(i[3,3])}")
                y.append( "    frame: body")
                y.append( "    inertia:")
                y.append(f"      Ixx: {str(i[0,0])}")
                y.append(f"      Ixy: {str(i[0,1])}")
                y.append(f"      Ixz: {str(i[0,2])}")
                y.append(f"      Iyy: {str(i[1,1])}")
                y.append(f"      Iyz: {str(i[1,2])}")
                y.append(f"      Izz: {str(i[2,2])}")
                y.append(f"    com: [{str(skd.simplify(i[2,4]/i[3,3]))}, {str(skd.simplify(i[0,5]/i[3,3]))}, {str(skd.simplify(i[1,3]/i[3,3]))}]")
                y.append("")
    
    if skd.q is not None:
        y.append(f"q: {[str(i) for i in skd.q]}")    
    else:
        y.append(f"q: [{','.join(['q'+str(i+1) for i in range(dof)])}]")
    
    if skd.qd is not None:
        y.append(f"qd: {[str(i) for i in skd.qd]}")    
    else:
        y.append(f"qd: [{','.join(['dq'+str(i+1) for i in range(dof)])}]")
    
    if skd.q2d is not None:
        y.append(f"q2d: {[str(i) for i in skd.q2d]}")    
    else:
        y.append(f"q2d: [{','.join(['ddq'+str(i+1) for i in range(dof)])}]")
    
    if skd.q3d is not None:
        y.append(f"q3d: {[str(i) for i in skd.q3d]}")    
    
    if skd.q4d is not None:
        y.append(f"q4d: {[str(i) for i in skd.q4d]}")
    y.append("")
    
    
    if type(skd.WEE) is list:
        y.append("WEE:")
        for i in skd.WEE:
            y.append(f"  - {[str(j) for j in i]}")
    else: 
        y.append(f"WEE: {[str(i) for i in skd.WEE]}")
    
    if type(skd.WDEE) is list:
        y.append("WDEE:")
        for i in skd.WDEE:
            y.append(f"  - {[str(j) for j in i]}")
    elif skd.WDEE != zeros(6,1): 
        y.append(f"WDEE: {[str(i) for i in skd.WDEE]}")
    
    if type(skd.W2DEE) is list:
        y.append("W2DEE:")
        for i in skd.W2DEE:
            y.append(f"  - {[str(j) for j in i]}")
    elif skd.W2DEE != zeros(6,1): 
        y.append(f"W2DEE: {[str(i) for i in skd.W2DEE]}")

    y.append("")
    
    if skd.assignment_dict:
        y.append("assignments:")
        for i in skd.assignment_dict:
            y.append(f"  {i}: {skd.assignment_dict[i]}")
        y.append("")
    
    for i in range(len(y)):
        y[i] = y[i].replace("'","")
        # remove trailing zeros
        y[i] = regex.sub("(?<=\W\d+\.\d+)0+(?=(\D|$))|(?<=\W\d+)\.0+(?=(\D|$))", "",y[i])
    if "return_dict" in kwargs and kwargs["return_dict"]:
        return yaml.load("\n".join(y))
        
    
    # check if path ends with .yaml
    if not bool(regex.search("(.yaml|.YAML|.yml|.YML)\Z", path)):
        path += ".yaml"
        
    with open(path, "w+") as f:
        f.write("\n".join(y))
        


def skd_to_json(skd: SymbolicKinDyn, path: str="robot.json") -> None:
    """Generate JSON file from SymbolicKinDyn object.

    Args:
        skd (SymbolicKinDyn): SymbolicKinDyn with set parameters.
        path (str, optional): path where to save .json file. 
            Defaults to "robot.json".
    """
    d = skd_to_yaml(skd,return_dict=True)
    s= json.dumps(d,indent=2)
    # check if path ends with .json
    if not bool(regex.search("(.JSON|.json)\Z", path)):
        path += ".json"
    
    with open(path, "w+") as f:
        f.write(s)

def urdf_to_yaml(urdf_path: str, yaml_path: str="robot.yaml", 
                 symbolic: bool=False, cse: bool=False, 
                 simplify_numbers: bool=True, tolerance: float=0.0001, 
                 max_denominator: int=9) -> SymbolicKinDyn:
    """Load robot from urdf and save it to YAML file.

        Args:
            urdf_path (str): path to URDF.
            yaml_path (str, optional): path where to save .yaml file. 
                Defaults to "robot.yaml".
            symbolic (bool, optional): 
                generate symbols for numeric values. 
                Defaults to False.
            cse (bool, optional): 
                use common subexpression elimination. Defaults to False.
            simplify_numbers (bool, optional): 
                Use eg. pi/2 instead of 1.5708. Defaults to True.
            tolerance (float, optional): 
                tolerance for simplify_numbers. Defaults to 0.0001.
            max_denominator (int, optional): 
                Maximum denominator to use for simplify numbers to avoid
                values like 13/153. Use 0 to deactivate. Defaults to 9.

        Raises:
            NotImplementedError: supports only the joint types 
                "revolute", "continuous" and "prismatic".
    """            
    skd = SymbolicKinDyn()
    skd.load_from_urdf(path=urdf_path, symbolic=symbolic, cse=cse, simplify_numbers=simplify_numbers, 
                       tolerance=tolerance, max_denominator=max_denominator)
    return skd_to_yaml(skd=skd,path=yaml_path)

def urdf_to_json(urdf_path: str, json_path: str="robot.json", 
                 symbolic: bool=False, cse: bool=False, 
                 simplify_numbers: bool=True, tolerance: float=0.0001, 
                 max_denominator: int=9) -> SymbolicKinDyn:
    """Load robot from urdf and save it to JSON file.

        Args:
            urdf_path (str): path to URDF.
            json_path (str, optional): path where to save .json file. 
                Defaults to "robot.json".
            symbolic (bool, optional): 
                generate symbols for numeric values. 
                Defaults to False.
            cse (bool, optional): 
                use common subexpression elimination. Defaults to False.
            simplify_numbers (bool, optional): 
                Use eg. pi/2 instead of 1.5708. Defaults to True.
            tolerance (float, optional): 
                tolerance for simplify_numbers. Defaults to 0.0001.
            max_denominator (int, optional): 
                Maximum denominator to use for simplify numbers to avoid
                values like 13/153. Use 0 to deactivate. Defaults to 9.

        Raises:
            NotImplementedError: supports only the joint types 
                "revolute", "continuous" and "prismatic".
    """            
    skd = SymbolicKinDyn()
    skd.load_from_urdf(path=urdf_path, symbolic=symbolic, cse=cse, 
                       simplify_numbers=simplify_numbers, tolerance=tolerance, 
                       max_denominator=max_denominator)
    return skd_to_json(skd=skd,path=json_path)


from sympy import symbols, Matrix, Identity, init_printing
from skidy import SymbolicKinDyn, mass_matrix_mixed_data, transformation_matrix, inertia_matrix


init_printing()

skd = SymbolicKinDyn()

# Declaration of symbolic variables
q0, q1, q2 = symbols("q0 q1 q2")
dq0, dq1, dq2 = symbols("dq0 dq1 dq2")
ddq0, ddq1, ddq2 = symbols("ddq0 ddq1 ddq2")
g, L1, L2, Ixx1, Ixy1, Ixz1, Iyy1, Iyz1, Izz1, Ixx2, Ixy2, Ixz2, Iyy2, Iyz2, Izz2, m0, m1, m2, cx1, cy1, cz1, cx2, cy2, cz2 = symbols(
    "g L1 L2 Ixx1 Ixy1 Ixz1 Iyy1 Iyz1 Izz1 Ixx2 Ixy2 Ixz2 Iyy2 Iyz2 Izz2 m0 m1 m2 cx1 cy2 cz1 cx2 cy2 cz2", constant=True, real=True)

skd.gravity_vector = Matrix([-g, 0, 0])

# Joint screw coordinates in spatial representation
# define config_representation to be spatial
skd.config_representation = skd.SPATIAL

skd.joint_screw_coord = []
skd.joint_screw_coord.append(Matrix([0, 0, 0, -1, 0, 0])) # prismatic joint

e1 = Matrix([0, 0, 1])  # joint axis of revolute joint
y1 = Matrix([0, 0, 0])  # Vector to joint axis from inertial Frame
# Joint screw coordinates in spatial representation
skd.joint_screw_coord.append(Matrix([e1, y1.cross(e1)])) # revolute joint 1

e2 = Matrix([0, 0, 1])  # joint axis of revolute joint
y2 = Matrix([L1, 0, 0])  # Vector to joint axis from inertial Frame
# Joint screw coordinates in spatial representation
skd.joint_screw_coord.append(Matrix([e2, y2.cross(e2)])) # revolute joint 2

# Reference configurations of bodies (i.e. of spatial frames)
r0 = Matrix([0, 0, 0])
r1 = Matrix([0, 0, 0])
r2 = Matrix([L1, 0, 0])

skd.body_ref_config = []
skd.body_ref_config.append(transformation_matrix(t=r0))
skd.body_ref_config.append(transformation_matrix(t=r1))
skd.body_ref_config.append(transformation_matrix(t=r2))

# Reference configurations of bodies (i.e. of body-fixed reference frames)
# define config_representation to be body fixed
# s.config_representation = s.BODY_FIXED
# r0 = Matrix([0, 0, 0]) 
# r1 = Matrix([0, 0, 0]) 
# r2 = Matrix([L1, 0, 0])

# s.body_ref_config = []
# s.body_ref_config.append(transformation_matrix(t=r0))
# s.body_ref_config.append(transformation_matrix(t=r1))
# s.body_ref_config.append(transformation_matrix(t=r2))



# End-effector configuration wrt last link body fixed frame in the chain
re = Matrix([L2, 0, 0])
skd.ee = transformation_matrix(t=re)

# Joint screw coordinates in body-fixed representation computed from screw coordinates in IFR
# s.joint_screw_coord = []
# s.joint_screw_coord.append(s.SE3AdjInvMatrix(s.A[0])*s.Y[0])
# s.joint_screw_coord.append(s.SE3AdjInvMatrix(s.A[1])*s.Y[1])
# s.joint_screw_coord.append(s.SE3AdjInvMatrix(s.A[2])*s.Y[2])



# Joint screw coordinates in body-fixed representation
# s.joint_screw_coord = []
# s.joint_screw_coord.append(Matrix([0,0,0,-1,0,0]))
# s.joint_screw_coord.append(Matrix([0,0,1,0,0,0]))
# s.joint_screw_coord.append(Matrix([0,0,1,0,0,0]))

# Mass-Inertia parameters

skd.Mb = []
skd.Mb.append(mass_matrix_mixed_data(
    m0, Matrix(Identity(3)), Matrix([0, 0, 0])))
skd.Mb.append(mass_matrix_mixed_data(m1, inertia_matrix(
    Ixx1, Ixy1, Ixz1, Iyy1, Ixz1, Izz1), Matrix([cx1, cy1, cz1])))
skd.Mb.append(mass_matrix_mixed_data(m2, inertia_matrix(
    Ixx2, Ixy2, Ixz2, Iyy2, Ixz2, Izz2), Matrix([cx2, cy2, cz2])))

# Declaring generalized vectors
q = Matrix([q0, q1, q2])
qd = Matrix([dq0, dq1, dq2])
q2d = Matrix([ddq0, ddq1, ddq2])

# Kinematics
F = skd.closed_form_kinematics_body_fixed(q, qd, q2d)
Q = skd.closed_form_inv_dyn_body_fixed(q, qd, q2d)
skd.generate_code(python=True, C=True, Matlab=True,
                use_global_vars=True, name="hopperPlant", project="HoppingLeg")



from skidy import (SymbolicKinDyn,
                   transformation_matrix,
                   mass_matrix_mixed_data,
                   joint_screw,
                   inertia_matrix,
                   SO3Exp,
                   generalized_vectors)
from skidy.symbols import g, L1, m1
import sympy

# Define connectivity graph
parent = [0]

child = [[]]

support = [[1]]

# gravity vector
gravity = sympy.Matrix([0,-g,0])

# representation of joint screw coordinates and body reference configurations
representation = 'body_fixed'

# joint screw coordinates (6x1 sympy.Matrix per joint)
joint_screw_coord = []
joint_screw_coord.append(joint_screw(axis=[0,0,1], vec=[0,0,0], revolute=True))

# body reference configurations (4x4 SE3 Pose (sympy.Matrix) per link)
body_ref_config = []
body_ref_config.append(transformation_matrix(r=SO3Exp(axis=[0,0,1],angle=0),t=[0,0,0]))

# end-effector configuration w.r.t. last link body fixed frame in the chain (4x4 SE3 Pose (sympy.Matrix))
ee = transformation_matrix(r=SO3Exp(axis=[0,0,1],angle=0),t=[0,-L1,0])

# mass_inertia parameters (6x6 sympy.Matrix per link)
I1 = m1*L1**2

Mb = []
Mb.append(mass_matrix_mixed_data(m1, I1*sympy.Identity(3) , sympy.Matrix([0,-L1,0])))

q, qd, q2d = generalized_vectors(len(body_ref_config), startindex=1)
WEE = sympy.zeros(6,1)

skd = SymbolicKinDyn(gravity_vector=gravity,
                     ee=ee,
                     body_ref_config=body_ref_config,
                     joint_screw_coord=joint_screw_coord,
                     config_representation=representation,
                     Mb=Mb,
                     parent=parent,
                     child=child,
                     support=support,
                     )

# run Calculations
skd.closed_form_kinematics_body_fixed(q, qd, q2d, simplify=True)
skd.closed_form_inv_dyn_body_fixed(q, qd, q2d, WEE=WEE, simplify=True)

# Generate Code
skd.generate_code(python=True, C=False, Matlab=False, latex=True,
                 folder="./generated_code", use_global_vars=True,
                 name="SinglePendulumPlant", project="SinglePendulum")



#!/usr/bin/env python3

import os
import sys
import argparse
from skidy.parser import (
    robot_from_yaml, generate_template_yaml,
    robot_from_json, generate_template_json,
    robot_from_urdf, generate_template_python,
    urdf_to_yaml, urdf_to_json 
)
from skidy import __version__

def main() -> None:
    parser = argparse.ArgumentParser(
        prog="python -m skidy",
        description="Symbolic kinematics and dynamics model generation using Equations of Motion in closed form.",
    )
    # TODO: add urdf to description as soon as supported.
    parser.add_argument("filename",type=str, help=".yaml or .json file with definition of robot to analyse or location where to save generated template with '--template' option.")
    parser.add_argument('--version', action='version',
                    version='skidy {version}'.format(version=__version__))
    parser.add_argument("-s", "--simplify", action="store_true", help="simplify expressions")
    parser.add_argument("--cse", action="store_true", help="use common subexpression elimination to shorten expressions")
    parser.add_argument("-p", "--python", action="store_true", help="generate python code")
    parser.add_argument("-c","--cpp", action="store_true", help="generate C++ code")
    parser.add_argument("-C","--C", action="store_true", help="generate C code")
    parser.add_argument("-m","--matlab", action="store_true", help="generate Matlab/Octave code")
    parser.add_argument("-j","--julia", action="store_true", help="generate julia code")
    parser.add_argument("-l","--latex", action="store_true", help="generate LaTeX code and pdf")
    parser.add_argument("-g","--graph", action="store_true", help="generate structure graph from input and save it as pdf")
    parser.add_argument("--cython", action="store_true", help="generate Cython code")
    parser.add_argument("--no-kinematics", action="store_true", help="skip generation of kinematics equations.")
    parser.add_argument("--no-dynamics", action="store_true", help="skip generation of dynamics equations.")
    parser.add_argument("-f", "--folder", type=str, default="./generated_code", help="folder where to save generated code. Defaults to './generated_code'")
    parser.add_argument("--serial", "--not-parallel", action="store_false", default=True, help="don't use parallel computation")
    parser.add_argument("-n","--name", type=str, default="", help="name of class and file. Defaults to filename")
    parser.add_argument("--project", type=str, default="Project", help="project name in C header. Defaults to 'Project'.")
    parser.add_argument("--cache", action="store_true", help="Use LRU cache for sin and cos function in generated python and cython code to speed up calculations.")
    # parser.add_argument("--yaml",action="store_true", help="enforce yaml file")
    # parser.add_argument("--json",action="store_true", help="enforce json file")
    # parser.add_argument("--urdf",action="store_true", help="enforce urdf file")

    # options for template generation
    generate_template = parser.add_argument_group("Options for template Generation (available formats: yaml, json, python)")
    generate_template.add_argument("-T","--template", "--please", action="store_true", help="store template yaml or json file to edit instead of analyzing existing robot. Format is chosen by extension of filename")
    generate_template.add_argument("-S", "--structure", type=str, help="structure of robot template: String containing only 'r' and 'p' of joint order. E.g.: Use 'prr' for a robot which has 1 prismatic joint followed by 2 revolute joints.")
    generate_template.add_argument("-d", "--dof", type=int, default=0, help="degrees of freedom. Is usually calculated by length of 'structure'.")
    generate_template.add_argument("--urdf",action="store_true", help="use urdf in generated python file")
    generate_template.add_argument("--yaml",action="store_true", help="enforce yaml file generation")
    generate_template.add_argument("--json",action="store_true", help="enforce json file generation")
    generate_template.add_argument("--convert", metavar="URDF_PATH", type=str, default=None, help="Convert URDF to YAML or JSON. Use -s option to round numbers if close to 0 or 1 (recommended)")
    generate_template.add_argument("--symbolify", action="store_true", help="Symbolify numbers during conversion. Defaults to False.")
    # generate_template.add_argument("--python", action="store_true", help="enforce python file generation")


    args = parser.parse_args()
    
    path = args.filename
    name, ext = os.path.splitext(path)
    
    # generate empty yaml or json file
    if args.convert or args.template:
        structure = args.structure
        dof = args.dof
        # yaml
        if args.yaml or ext in {".yaml",".YAML","yml"}:
            if not ext in {".yaml",".YAML","yml"}:
                path = name+".yaml"
            if (os.path.exists(path) 
                and not input(f"{path} already exists.\nOverwrite existing file? [Y/n] ") in {"","y","Y"}):
                print("Abort execution.")
                exit()
            if args.convert:
                urdf_to_yaml(args.convert,path,symbolic=args.symbolify, cse=args.cse,
                                  simplify_numbers=args.simplify,tolerance=0.0001, 
                                  max_denominator=9)
            else:
                generate_template_yaml(path,structure,dof)    
        # json
        if args.json or ext in {".json",".JSON"}:
            if ext not in {".json",".JSON"}:
                path = name+".json"
            if (os.path.exists(path) 
                and not input(f"{path} already exists.\nOverwrite existing file? [Y/n] ") in {"","y","Y"}):
                print("Abort execution.")
                exit()
            if args.convert:
                urdf_to_json(args.convert,path,symbolic=args.symbolify, cse=args.cse,
                                  simplify_numbers=args.simplify,tolerance=0.0001, 
                                  max_denominator=9)
            else:
                generate_template_json(path,structure,dof)
        # python
        if args.python or ext in {".py"}:
            if ext not in {".py"}:
                path = name+".py"
            if (os.path.exists(path) 
                and not input(f"{path} already exists.\nOverwrite existing file? [Y/n] ") in {"","y","Y"}):
                print("Abort execution.")
                exit()
            generate_template_python(path,structure,dof,urdf=args.urdf)
    else:
        if not os.path.exists(path) or not os.path.isfile(path):
            raise ValueError(f"{path} is no existing file.")
        
        if args.name == "": args.name = os.path.basename(name)
        
        if (not args.python 
            and not args.C 
            and not args.cpp
            and not args.matlab 
            and not args.julia 
            and not args.latex 
            and not args.cython
            and not args.graph):
            raise ValueError("Please provide at least one programming language in which I should generate code.")
        
        if ext in {".yaml",".YAML","yml"}:
            skd = robot_from_yaml(path)
            
        elif ext in {".json",".JSON"}:
            skd = robot_from_json(path)
            
        elif ext in {".urdf"}:
            skd = robot_from_urdf(path,symbolic=False, cse=args.cse,
                                  simplify_numbers=True,tolerance=0.0001, 
                                  max_denominator=8)
        
        # elif ext in {".urdf"}:
        #     skd = robot_from_urdf(path,cse=args.cse) # TODO: add options
        #     # TODO: deal with ee

        elif ext == ".py":
            if input(f"{path} is a python file. Should I run this file? [y/N] ") in {"y","Y"}:
                os.system(f"{sys.executable} {path}")
            else:
                print("Abort execution.")
            exit()
        else: 
            raise ValueError("File extension not recognized.")
        
        if (args.python 
            or args.C 
            or args.cpp
            or args.matlab 
            or args.julia 
            or args.latex 
            or args.cython):
            
            if not args.no_kinematics:
                skd.closed_form_kinematics_body_fixed(
                    simplify=args.simplify, cse=args.cse, parallel=args.serial)
            if not args.no_dynamics:
                skd.closed_form_inv_dyn_body_fixed(
                    simplify=args.simplify, cse=args.cse, parallel=args.serial)
    
            skd.generate_code(python=args.python, 
                              C=args.C, 
                              cpp=args.cpp,
                              Matlab=args.matlab, 
                              cython=args.cython, 
                              julia=args.julia, 
                              latex=args.latex, 
                              cse=args.cse,
                              cache=args.cache,
                              folder=args.folder, 
                              use_global_vars=True, 
                              name=args.name, 
                              project=args.project)
        if args.graph:
            if not os.path.exists(os.path.join(args.folder,"graphs")):
                os.mkdir(os.path.join(args.folder,"graphs"))
            skd.generate_graph(os.path.join(args.folder,"graphs",args.name+".pdf"),1)

if __name__ == "__main__":
    main()


from skidy import SymbolicKinDyn, mass_matrix_mixed_data, transformation_matrix
from sympy import symbols, Matrix, Identity, init_printing
import sys
from os.path import dirname
sys.path.append(dirname(dirname(__file__)))


init_printing()

# Declaration of symbolic variables
q1, q2 = symbols("q1 q2")  # joint positions
dq1, dq2 = symbols("dq1 dq2")  # joint velocities
ddq1, ddq2 = symbols("ddq1 ddq2")  # joint accelerations

parent = [0,
          1]

child = [[1],
         []]

support = [[1],
           [1,2]]


# mass and inertia values
m1, m2 = symbols("m1 m2", real=1, constant=1)

# gravity
g = symbols("g", real=1, constant=1)
L1, L2 = symbols("L1 L2", real=1, constant=1)  # link lengths
pi = symbols("pi", real=1, constant=1)  # pi

gravity_vector = Matrix([0, -g, 0])  # define gravity vector

# Joint screw coordinates in spatial representation

joint_screw_coord = []
e1 = Matrix([0, 0, 1])  # joint axis of revolute joint
y1 = Matrix([0, 0, 0])  # Vector to joint axis from inertial Frame
# Joint screw coordinates in spatial representation (6,1)
joint_screw_coord.append(Matrix([e1, y1.cross(e1)]))

e2 = Matrix([0, 0, 1])  # joint axis of revolute joint
y2 = Matrix([0, -L1, 0])  # Vector to joint axis from inertial Frame
# Joint screw coordinates in spatial representation (6,1)
joint_screw_coord.append(Matrix([e2, y2.cross(e2)]))



# Reference configurations of bodies (i.e. of absolute reference frames)
r1 = Matrix([0, 0, 0])
r2 = Matrix([0, -L1, 0])

body_ref_config = []
# no rotation, just translation
body_ref_config.append(transformation_matrix(t=r1))
body_ref_config.append(transformation_matrix(t=r2))

# End-effector configuration wrt last link body fixed frame in the chain
re = Matrix([0, -L2, 0])
ee = transformation_matrix(t=re)

# Mass-Inertia parameters
cg1 = Matrix([0, -L1, 0]).T
cg2 = Matrix([0, -L2, 0]).T
I1 = m1*L1**2
I2 = m2*L2**2

Mb = []
Mb.append(mass_matrix_mixed_data(m1, I1*Identity(3), cg1))
Mb.append(mass_matrix_mixed_data(m2, I2*Identity(3), cg2))

# Declaring generalized vectors
q = Matrix([q1, q2])
qd = Matrix([dq1, dq2])
q2d = Matrix([ddq1, ddq2])

s = SymbolicKinDyn(gravity_vector=gravity_vector, 
                    ee=ee,
                    body_ref_config=body_ref_config,
                    joint_screw_coord=joint_screw_coord,
                    config_representation="spatial", 
                    Mb=Mb, parent=parent, child=child, support=support)
s.closed_form_kinematics_body_fixed(q,qd,q2d)
s.closed_form_inv_dyn_body_fixed(q,qd,q2d)
s.generate_code(python=True, C=False, Matlab=False, latex=False,name="DoublePendulumPlant",project="DoublePendulum")
# s.generate_graph("generated_code/graphs/DoublePendulum.pdf")


"""symbols:
This module exports all latin and greek letters, as well as some common 
used letter combinations with and without indices between 0 and 10.

Usage:
    To use the defined symbols use e.g.:
    
    `from skidy.symbols import m1, l1, Ixx1, Fx`

"""
from sympy import symbols
from sympy.abc import *

# q0, q1, q2, q3, q4, q5, q6, q7, q8, q9, q10 = symbols("q0, q1, q2, q3, q4, q5, q6, q7, q8, q9, q10", real=1, constant = 0)
dq, dq0, dq1, dq2, dq3, dq4, dq5, dq6, dq7, dq8, dq9, dq10 = symbols("dq, dq0, dq1, dq2, dq3, dq4, dq5, dq6, dq7, dq8, dq9, dq10", real=1, constant = 0)
ddq, ddq0, ddq1, ddq2, ddq3, ddq4, ddq5, ddq6, ddq7, ddq8, ddq9, ddq10 = symbols("ddq, ddq0, ddq1, ddq2, ddq3, ddq4, ddq5, ddq6, ddq7, ddq8, ddq9, ddq10", real=1, constant = 0)
dddq, dddq0, dddq1, dddq2, dddq3, dddq4, dddq5, dddq6, dddq7, dddq8, dddq9, dddq10 = symbols("dddq, dddq0, dddq1, dddq2, dddq3, dddq4, dddq5, dddq6, dddq7, dddq8, dddq9, dddq10", real=1, constant = 0)
ddddq, ddddq0, ddddq1, ddddq2, ddddq3, ddddq4, ddddq5, ddddq6, ddddq7, ddddq8, ddddq9, ddddq10 = symbols("ddddq, ddddq0, ddddq1, ddddq2, ddddq3, ddddq4, ddddq5, ddddq6, ddddq7, ddddq8, ddddq9, ddddq10", real=1, constant = 0)

Fx, Fy, Fz, Mx, My, Mz, fx, fy, fz, mx, my, mz, taux, tauy, tauz = symbols("Fx, Fy, Fz, Mx, My, Mz, fx, fy, fz, mx, my, mz, taux, tauy, tauz", real=1)
dFx, dFy, dFz, dMx, dMy, dMz, dfx, dfy, dfz, dmx, dmy, dmz, dtaux, dtauy, dtauz = symbols("dFx, dFy, dFz, dMx, dMy, dMz, dfx, dfy, dfz, dmx, dmy, dmz, dtaux, dtauy, dtauz", real=1)
ddFx, ddFy, ddFz, ddMx, ddMy, ddMz, ddfx, ddfy, ddfz, ddmx, ddmy, ddmz, ddtaux, ddtauy, ddtauz = symbols("ddFx, ddFy, ddFz, ddMx, ddMy, ddMz, ddfx, ddfy, ddfz, ddmx, ddmy, ddmz, ddtaux, ddtauy, ddtauz", real=1)
dF, dM, df, dm, dtau, ddF, ddM, ddf, ddm, ddtau = symbols("dF, dM, df, dm, dtau, ddF, ddM, ddf, ddm, ddtau", real=1)  

# m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10 = symbols("m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10", real=1, constant = 1)

# l0, l1, l2, l3, l4, l5, l6, l7, l8, l9, l10 = symbols("l0, l1, l2, l3, l4, l5, l6, l7, l8, l9, l10", real=1, constant = 1)
# L0, L1, L2, L3, L4, L5, L6, L7, L8, L9, L10 = symbols("L0, L1, L2, L3, L4, L5, L6, L7, L8, L9, L10", real=1, constant = 1)

# x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10 = symbols("x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10", real=1, constant = 1)
# y0, y1, y2, y3, y4, y5, y6, y7, y8, y9, y10 = symbols("y0, y1, y2, y3, y4, y5, y6, y7, y8, y9, y10", real=1, constant = 1)
# z0, z1, z2, z3, z4, z5, z6, z7, z8, z9, z10 = symbols("z0, z1, z2, z3, z4, z5, z6, z7, z8, z9, z10", real=1, constant = 1)

# X0, X1, X2, X3, X4, X5, X6, X7, X8, X9, X10 = symbols("X0, X1, X2, X3, X4, X5, X6, X7, X8, X9, X10", real=1, constant = 1)
# Y0, Y1, Y2, Y3, Y4, Y5, Y6, Y7, Y8, Y9, Y10 = symbols("Y0, Y1, Y2, Y3, Y4, Y5, Y6, Y7, Y8, Y9, Y10", real=1, constant = 1)
# Z0, Z1, Z2, Z3, Z4, Z5, Z6, Z7, Z8, Z9, Z10 = symbols("Z0, Z1, Z2, Z3, Z4, Z5, Z6, Z7, Z8, Z9, Z10", real=1, constant = 1)

# I0, I1, I2, I3, I4, I5, I6, I7, I8, I9, I10 = symbols("I0, I1, I2, I3, I4, I5, I6, I7, I8, I9, I10", real=1, constant = 1)
# i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10 = symbols("i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10", real=1, constant = 1)

Ixx, Ixx0, Ixx1, Ixx2, Ixx3, Ixx4, Ixx5, Ixx6, Ixx7, Ixx8, Ixx9, Ixx10 = symbols("Ixx, Ixx0, Ixx1, Ixx2, Ixx3, Ixx4, Ixx5, Ixx6, Ixx7, Ixx8, Ixx9, Ixx10", real=1, constant = 1)
Iyy, Iyy0, Iyy1, Iyy2, Iyy3, Iyy4, Iyy5, Iyy6, Iyy7, Iyy8, Iyy9, Iyy10 = symbols("Iyy, Iyy0, Iyy1, Iyy2, Iyy3, Iyy4, Iyy5, Iyy6, Iyy7, Iyy8, Iyy9, Iyy10", real=1, constant = 1)
Izz, Izz0, Izz1, Izz2, Izz3, Izz4, Izz5, Izz6, Izz7, Izz8, Izz9, Izz10 = symbols("Izz, Izz0, Izz1, Izz2, Izz3, Izz4, Izz5, Izz6, Izz7, Izz8, Izz9, Izz10", real=1, constant = 1)
Ixy, Ixy0, Ixy1, Ixy2, Ixy3, Ixy4, Ixy5, Ixy6, Ixy7, Ixy8, Ixy9, Ixy10 = symbols("Ixy, Ixy0, Ixy1, Ixy2, Ixy3, Ixy4, Ixy5, Ixy6, Ixy7, Ixy8, Ixy9, Ixy10", real=1, constant = 1)
Ixz, Ixz0, Ixz1, Ixz2, Ixz3, Ixz4, Ixz5, Ixz6, Ixz7, Ixz8, Ixz9, Ixz10 = symbols("Ixz, Ixz0, Ixz1, Ixz2, Ixz3, Ixz4, Ixz5, Ixz6, Ixz7, Ixz8, Ixz9, Ixz10", real=1, constant = 1)
Iyz, Iyz0, Iyz1, Iyz2, Iyz3, Iyz4, Iyz5, Iyz6, Iyz7, Iyz8, Iyz9, Iyz10 = symbols("Iyz, Iyz0, Iyz1, Iyz2, Iyz3, Iyz4, Iyz5, Iyz6, Iyz7, Iyz8, Iyz9, Iyz10", real=1, constant = 1)

ixx, ixx0, ixx1, ixx2, ixx3, ixx4, ixx5, ixx6, ixx7, ixx8, ixx9, ixx10 = symbols("ixx, ixx0, ixx1, ixx2, ixx3, ixx4, ixx5, ixx6, ixx7, ixx8, ixx9, ixx10", real=1, constant = 1)
iyy, iyy0, iyy1, iyy2, iyy3, iyy4, iyy5, iyy6, iyy7, iyy8, iyy9, iyy10 = symbols("iyy, iyy0, iyy1, iyy2, iyy3, iyy4, iyy5, iyy6, iyy7, iyy8, iyy9, iyy10", real=1, constant = 1)
izz, izz0, izz1, izz2, izz3, izz4, izz5, izz6, izz7, izz8, izz9, izz10 = symbols("izz, izz0, izz1, izz2, izz3, izz4, izz5, izz6, izz7, izz8, izz9, izz10", real=1, constant = 1)
ixy, ixy0, ixy1, ixy2, ixy3, ixy4, ixy5, ixy6, ixy7, ixy8, ixy9, ixy10 = symbols("ixy, ixy0, ixy1, ixy2, ixy3, ixy4, ixy5, ixy6, ixy7, ixy8, ixy9, ixy10", real=1, constant = 1)
ixz, ixz0, ixz1, ixz2, ixz3, ixz4, ixz5, ixz6, ixz7, ixz8, ixz9, ixz10 = symbols("ixz, ixz0, ixz1, ixz2, ixz3, ixz4, ixz5, ixz6, ixz7, ixz8, ixz9, ixz10", real=1, constant = 1)
iyz, iyz0, iyz1, iyz2, iyz3, iyz4, iyz5, iyz6, iyz7, iyz8, iyz9, iyz10 = symbols("iyz, iyz0, iyz1, iyz2, iyz3, iyz4, iyz5, iyz6, iyz7, iyz8, iyz9, iyz10", real=1, constant = 1)

# c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10 = symbols("c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10", real=1, constant = 1)

cx, cx0, cx1, cx2, cx3, cx4, cx5, cx6, cx7, cx8, cx9, cx10 = symbols("cx, cx0, cx1, cx2, cx3, cx4, cx5, cx6, cx7, cx8, cx9, cx10", real=1, constant = 1)
cy, cy0, cy1, cy2, cy3, cy4, cy5, cy6, cy7, cy8, cy9, cy10 = symbols("cy, cy0, cy1, cy2, cy3, cy4, cy5, cy6, cy7, cy8, cy9, cy10", real=1, constant = 1)
cz, cz0, cz1, cz2, cz3, cz4, cz5, cz6, cz7, cz8, cz9, cz10 = symbols("cz, cz0, cz1, cz2, cz3, cz4, cz5, cz6, cz7, cz8, cz9, cz10", real=1, constant = 1)

cg, cg0, cg1, cg2, cg3, cg4, cg5, cg6, cg7, cg8, cg9, cg10 = symbols("cg, cg0, cg1, cg2, cg3, cg4, cg5, cg6, cg7, cg8, cg9, cg10", real=1, constant = 1)

cgx, cgx0, cgx1, cgx2, cgx3, cgx4, cgx5, cgx6, cgx7, cgx8, cgx9, cgx10 = symbols("cgx, cgx0, cgx1, cgx2, cgx3, cgx4, cgx5, cgx6, cgx7, cgx8, cgx9, cgx10", real=1, constant = 1)
cgy, cgy0, cgy1, cgy2, cgy3, cgy4, cgy5, cgy6, cgy7, cgy8, cgy9, cgy10 = symbols("cgy, cgy0, cgy1, cgy2, cgy3, cgy4, cgy5, cgy6, cgy7, cgy8, cgy9, cgy10", real=1, constant = 1)
cgz, cgz0, cgz1, cgz2, cgz3, cgz4, cgz5, cgz6, cgz7, cgz8, cgz9, cgz10 = symbols("cgz, cgz0, cgz1, cgz2, cgz3, cgz4, cgz5, cgz6, cgz7, cgz8, cgz9, cgz10", real=1, constant = 1)

Lc, Lc0, Lc1, Lc2, Lc3, Lc4, Lc5, Lc6, Lc7, Lc8, Lc9, Lc10 = symbols("Lc, Lc0, Lc1, Lc2, Lc3, Lc4, Lc5, Lc6, Lc7, Lc8, Lc9, Lc10", real=1, constant = 1)
lc, lc0, lc1, lc2, lc3, lc4, lc5, lc6, lc7, lc8, lc9, lc10 = symbols("lc, lc0, lc1, lc2, lc3, lc4, lc5, lc6, lc7, lc8, lc9, lc10", real=1, constant = 1)

a0, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10 = symbols("a0, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10", real=True)
b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10 = symbols("b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10", real=True)
c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10 = symbols("c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10", real=True)
d0, d1, d2, d3, d4, d5, d6, d7, d8, d9, d10 = symbols("d0, d1, d2, d3, d4, d5, d6, d7, d8, d9, d10", real=True)
e0, e1, e2, e3, e4, e5, e6, e7, e8, e9, e10 = symbols("e0, e1, e2, e3, e4, e5, e6, e7, e8, e9, e10", real=True)
f0, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10 = symbols("f0, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10", real=True)
g0, g1, g2, g3, g4, g5, g6, g7, g8, g9, g10 = symbols("g0, g1, g2, g3, g4, g5, g6, g7, g8, g9, g10", real=True)
h0, h1, h2, h3, h4, h5, h6, h7, h8, h9, h10 = symbols("h0, h1, h2, h3, h4, h5, h6, h7, h8, h9, h10", real=True)
i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10 = symbols("i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10", real=True)
j0, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10 = symbols("j0, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10", real=True)
k0, k1, k2, k3, k4, k5, k6, k7, k8, k9, k10 = symbols("k0, k1, k2, k3, k4, k5, k6, k7, k8, k9, k10", real=True)
l0, l1, l2, l3, l4, l5, l6, l7, l8, l9, l10 = symbols("l0, l1, l2, l3, l4, l5, l6, l7, l8, l9, l10", real=True)
m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10 = symbols("m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10", real=True)
n0, n1, n2, n3, n4, n5, n6, n7, n8, n9, n10 = symbols("n0, n1, n2, n3, n4, n5, n6, n7, n8, n9, n10", real=True)
o0, o1, o2, o3, o4, o5, o6, o7, o8, o9, o10 = symbols("o0, o1, o2, o3, o4, o5, o6, o7, o8, o9, o10", real=True)
p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10 = symbols("p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10", real=True)
q0, q1, q2, q3, q4, q5, q6, q7, q8, q9, q10 = symbols("q0, q1, q2, q3, q4, q5, q6, q7, q8, q9, q10", real=True)
u0, u1, u2, u3, u4, u5, u6, u7, u8, u9, u10 = symbols("u0, u1, u2, u3, u4, u5, u6, u7, u8, u9, u10", real=True)
r0, r1, r2, r3, r4, r5, r6, r7, r8, r9, r10 = symbols("r0, r1, r2, r3, r4, r5, r6, r7, r8, r9, r10", real=True)
s0, s1, s2, s3, s4, s5, s6, s7, s8, s9, s10 = symbols("s0, s1, s2, s3, s4, s5, s6, s7, s8, s9, s10", real=True)
t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10 = symbols("t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10", real=True)
u0, u1, u2, u3, u4, u5, u6, u7, u8, u9, u10 = symbols("u0, u1, u2, u3, u4, u5, u6, u7, u8, u9, u10", real=True)
v0, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10 = symbols("v0, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10", real=True)
w0, w1, w2, w3, w4, w5, w6, w7, w8, w9, w10 = symbols("w0, w1, w2, w3, w4, w5, w6, w7, w8, w9, w10", real=True)
x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10 = symbols("x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10", real=True)
y0, y1, y2, y3, y4, y5, y6, y7, y8, y9, y10 = symbols("y0, y1, y2, y3, y4, y5, y6, y7, y8, y9, y10", real=True)
z0, z1, z2, z3, z4, z5, z6, z7, z8, z9, z10 = symbols("z0, z1, z2, z3, z4, z5, z6, z7, z8, z9, z10", real=True)
A0, A1, A2, A3, A4, A5, A6, A7, A8, A9, A10 = symbols("A0, A1, A2, A3, A4, A5, A6, A7, A8, A9, A10", real=True)
B0, B1, B2, B3, B4, B5, B6, B7, B8, B9, B10 = symbols("B0, B1, B2, B3, B4, B5, B6, B7, B8, B9, B10", real=True)
C0, C1, C2, C3, C4, C5, C6, C7, C8, C9, C10 = symbols("C0, C1, C2, C3, C4, C5, C6, C7, C8, C9, C10", real=True)
D0, D1, D2, D3, D4, D5, D6, D7, D8, D9, D10 = symbols("D0, D1, D2, D3, D4, D5, D6, D7, D8, D9, D10", real=True)
E0, E1, E2, E3, E4, E5, E6, E7, E8, E9, E10 = symbols("E0, E1, E2, E3, E4, E5, E6, E7, E8, E9, E10", real=True)
F0, F1, F2, F3, F4, F5, F6, F7, F8, F9, F10 = symbols("F0, F1, F2, F3, F4, F5, F6, F7, F8, F9, F10", real=True)
G0, G1, G2, G3, G4, G5, G6, G7, G8, G9, G10 = symbols("G0, G1, G2, G3, G4, G5, G6, G7, G8, G9, G10", real=True)
H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10 = symbols("H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10", real=True)
I0, I1, I2, I3, I4, I5, I6, I7, I8, I9, I10 = symbols("I0, I1, I2, I3, I4, I5, I6, I7, I8, I9, I10", real=True)
J0, J1, J2, J3, J4, J5, J6, J7, J8, J9, J10 = symbols("J0, J1, J2, J3, J4, J5, J6, J7, J8, J9, J10", real=True)
K0, K1, K2, K3, K4, K5, K6, K7, K8, K9, K10 = symbols("K0, K1, K2, K3, K4, K5, K6, K7, K8, K9, K10", real=True)
L0, L1, L2, L3, L4, L5, L6, L7, L8, L9, L10 = symbols("L0, L1, L2, L3, L4, L5, L6, L7, L8, L9, L10", real=True)
M0, M1, M2, M3, M4, M5, M6, M7, M8, M9, M10 = symbols("M0, M1, M2, M3, M4, M5, M6, M7, M8, M9, M10", real=True)
N0, N1, N2, N3, N4, N5, N6, N7, N8, N9, N10 = symbols("N0, N1, N2, N3, N4, N5, N6, N7, N8, N9, N10", real=True)
O0, O1, O2, O3, O4, O5, O6, O7, O8, O9, O10 = symbols("O0, O1, O2, O3, O4, O5, O6, O7, O8, O9, O10", real=True)
P0, P1, P2, P3, P4, P5, P6, P7, P8, P9, P10 = symbols("P0, P1, P2, P3, P4, P5, P6, P7, P8, P9, P10", real=True)
Q0, Q1, Q2, Q3, Q4, Q5, Q6, Q7, Q8, Q9, Q10 = symbols("Q0, Q1, Q2, Q3, Q4, Q5, Q6, Q7, Q8, Q9, Q10", real=True)
R0, R1, R2, R3, R4, R5, R6, R7, R8, R9, R10 = symbols("R0, R1, R2, R3, R4, R5, R6, R7, R8, R9, R10", real=True)
S0, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10 = symbols("S0, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10", real=True)
T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10 = symbols("T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10", real=True)
U0, U1, U2, U3, U4, U5, U6, U7, U8, U9, U10 = symbols("U0, U1, U2, U3, U4, U5, U6, U7, U8, U9, U10", real=True)
V0, V1, V2, V3, V4, V5, V6, V7, V8, V9, V10 = symbols("V0, V1, V2, V3, V4, V5, V6, V7, V8, V9, V10", real=True)
W0, W1, W2, W3, W4, W5, W6, W7, W8, W9, W10 = symbols("W0, W1, W2, W3, W4, W5, W6, W7, W8, W9, W10", real=True)
X0, X1, X2, X3, X4, X5, X6, X7, X8, X9, X10 = symbols("X0, X1, X2, X3, X4, X5, X6, X7, X8, X9, X10", real=True)
Y0, Y1, Y2, Y3, Y4, Y5, Y6, Y7, Y8, Y9, Y10 = symbols("Y0, Y1, Y2, Y3, Y4, Y5, Y6, Y7, Y8, Y9, Y10", real=True)
Z0, Z1, Z2, Z3, Z4, Z5, Z6, Z7, Z8, Z9, Z10 = symbols("Z0, Z1, Z2, Z3, Z4, Z5, Z6, Z7, Z8, Z9, Z10", real=True)
alpha0, alpha1, alpha2, alpha3, alpha4, alpha5, alpha6, alpha7, alpha8, alpha9, alpha10 = symbols("alpha0, alpha1, alpha2, alpha3, alpha4, alpha5, alpha6, alpha7, alpha8, alpha9, alpha10", real=True)
beta0, beta1, beta2, beta3, beta4, beta5, beta6, beta7, beta8, beta9, beta10 = symbols("beta0, beta1, beta2, beta3, beta4, beta5, beta6, beta7, beta8, beta9, beta10", real=True)
gamma0, gamma1, gamma2, gamma3, gamma4, gamma5, gamma6, gamma7, gamma8, gamma9, gamma10 = symbols("gamma0, gamma1, gamma2, gamma3, gamma4, gamma5, gamma6, gamma7, gamma8, gamma9, gamma10", real=True)
delta0, delta1, delta2, delta3, delta4, delta5, delta6, delta7, delta8, delta9, delta10 = symbols("delta0, delta1, delta2, delta3, delta4, delta5, delta6, delta7, delta8, delta9, delta10", real=True)
epsilon0, epsilon1, epsilon2, epsilon3, epsilon4, epsilon5, epsilon6, epsilon7, epsilon8, epsilon9, epsilon10 = symbols("epsilon0, epsilon1, epsilon2, epsilon3, epsilon4, epsilon5, epsilon6, epsilon7, epsilon8, epsilon9, epsilon10", real=True)
zeta0, zeta1, zeta2, zeta3, zeta4, zeta5, zeta6, zeta7, zeta8, zeta9, zeta10 = symbols("zeta0, zeta1, zeta2, zeta3, zeta4, zeta5, zeta6, zeta7, zeta8, zeta9, zeta10", real=True)
eta0, eta1, eta2, eta3, eta4, eta5, eta6, eta7, eta8, eta9, eta10 = symbols("eta0, eta1, eta2, eta3, eta4, eta5, eta6, eta7, eta8, eta9, eta10", real=True)
theta0, theta1, theta2, theta3, theta4, theta5, theta6, theta7, theta8, theta9, theta10 = symbols("theta0, theta1, theta2, theta3, theta4, theta5, theta6, theta7, theta8, theta9, theta10", real=True)
iota0, iota1, iota2, iota3, iota4, iota5, iota6, iota7, iota8, iota9, iota10 = symbols("iota0, iota1, iota2, iota3, iota4, iota5, iota6, iota7, iota8, iota9, iota10", real=True)
kappa0, kappa1, kappa2, kappa3, kappa4, kappa5, kappa6, kappa7, kappa8, kappa9, kappa10 = symbols("kappa0, kappa1, kappa2, kappa3, kappa4, kappa5, kappa6, kappa7, kappa8, kappa9, kappa10", real=True)
lamda0, lamda1, lamda2, lamda3, lamda4, lamda5, lamda6, lamda7, lamda8, lamda9, lamda10 = symbols("lamda0, lamda1, lamda2, lamda3, lamda4, lamda5, lamda6, lamda7, lamda8, lamda9, lamda10", real=True)
mu0, mu1, mu2, mu3, mu4, mu5, mu6, mu7, mu8, mu9, mu10 = symbols("mu0, mu1, mu2, mu3, mu4, mu5, mu6, mu7, mu8, mu9, mu10", real=True)
nu0, nu1, nu2, nu3, nu4, nu5, nu6, nu7, nu8, nu9, nu10 = symbols("nu0, nu1, nu2, nu3, nu4, nu5, nu6, nu7, nu8, nu9, nu10", real=True)
xi0, xi1, xi2, xi3, xi4, xi5, xi6, xi7, xi8, xi9, xi10 = symbols("xi0, xi1, xi2, xi3, xi4, xi5, xi6, xi7, xi8, xi9, xi10", real=True)
omicron0, omicron1, omicron2, omicron3, omicron4, omicron5, omicron6, omicron7, omicron8, omicron9, omicron10 = symbols("omicron0, omicron1, omicron2, omicron3, omicron4, omicron5, omicron6, omicron7, omicron8, omicron9, omicron10", real=True)
pi0, pi1, pi2, pi3, pi4, pi5, pi6, pi7, pi8, pi9, pi10 = symbols("pi0, pi1, pi2, pi3, pi4, pi5, pi6, pi7, pi8, pi9, pi10", real=True)
rho0, rho1, rho2, rho3, rho4, rho5, rho6, rho7, rho8, rho9, rho10 = symbols("rho0, rho1, rho2, rho3, rho4, rho5, rho6, rho7, rho8, rho9, rho10", real=True)
sigma0, sigma1, sigma2, sigma3, sigma4, sigma5, sigma6, sigma7, sigma8, sigma9, sigma10 = symbols("sigma0, sigma1, sigma2, sigma3, sigma4, sigma5, sigma6, sigma7, sigma8, sigma9, sigma10", real=True)
tau0, tau1, tau2, tau3, tau4, tau5, tau6, tau7, tau8, tau9, tau10 = symbols("tau0, tau1, tau2, tau3, tau4, tau5, tau6, tau7, tau8, tau9, tau10", real=True)
upsilon0, upsilon1, upsilon2, upsilon3, upsilon4, upsilon5, upsilon6, upsilon7, upsilon8, upsilon9, upsilon10 = symbols("upsilon0, upsilon1, upsilon2, upsilon3, upsilon4, upsilon5, upsilon6, upsilon7, upsilon8, upsilon9, upsilon10", real=True)
phi0, phi1, phi2, phi3, phi4, phi5, phi6, phi7, phi8, phi9, phi10 = symbols("phi0, phi1, phi2, phi3, phi4, phi5, phi6, phi7, phi8, phi9, phi10", real=True)
chi0, chi1, chi2, chi3, chi4, chi5, chi6, chi7, chi8, chi9, chi10 = symbols("chi0, chi1, chi2, chi3, chi4, chi5, chi6, chi7, chi8, chi9, chi10", real=True)
psi0, psi1, psi2, psi3, psi4, psi5, psi6, psi7, psi8, psi9, psi10 = symbols("psi0, psi1, psi2, psi3, psi4, psi5, psi6, psi7, psi8, psi9, psi10", real=True)
omega0, omega1, omega2, omega3, omega4, omega5, omega6, omega7, omega8, omega9, omega10 = symbols("omega0, omega1, omega2, omega3, omega4, omega5, omega6, omega7, omega8, omega9, omega10", real=True)


from setuptools import setup, find_packages

with open("README.md", "r") as f:
        long_description = f.read()
        
setup(
    name="skidy",
    author="Underactuated Lab DFKI Robotics Innovation Center Bremen",
    maintainer="Hannah Isermann",
    maintainer_email="hannah.isermann@dfki.de",
    version="0.0.1",
    long_description=long_description,
    long_description_content_type='text/markdown',
    # url="",
    packages=find_packages("src",exclude="test"),
    package_dir={"": "src"},
    setup_requires=["numpy"],
    python_requires=">=3.8",
    install_requires=[
        "numpy",
        "sympy>=1.8",
        "urdf_parser_py>=0.0.4",
        "regex",
        "PyYAML",
        "pylatex",
        "pydot",
    ],
    extras_require={"testing": ["cython","oct2py","kinpy","pin"]},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Programming Language :: Python',
        "Operating System :: OS Independent",
    ],
)